from __future__ import division
from builtins import object

import sys, os, copy
from psychopy import visual, monitors, prefs, constants
from psychopy.visual import filters
from psychopy.tools.coordinatetools import pol2cart
from psychopy.tests import utils
import numpy
import pytest
import shutil
from tempfile import mkdtemp

"""Each test class creates a context subclasses _baseVisualTest to run a series
of tests on a single graphics context (e.g. pyglet with shaders)

To add a new stimulus test use _base so that it gets tested in all contexts

"""

# are we testing on Travis and is it Anaconda or system python?
_travisTesting = bool("{}".format(os.environ.get('TRAVIS')).lower() == 'true')
_anacondaTesting = bool("{}".format(os.environ.get('CONDA')).lower() == 'true')
# the ffmpeg doesn't seem to work on Travis system python (using 12.04)
# upgrading to trusty (14.04) we could get ffmpeg to work but then test_bitsShaders
# stopped working on conda and system python setup would even build with all the
# dependencies. It was test environment hell! (sorry, it's been a bad day)

class Test_Window(object):
    """Some tests just for the window - we don't really care about what's drawn inside it
    """
    def setup_class(self):
        self.temp_dir = mkdtemp(prefix='psychopy-tests-test_window')
        self.win = visual.Window([128,128], pos=[50,50], allowGUI=False, autoLog=False)

    def teardown_class(self):
        shutil.rmtree(self.temp_dir)

    def test_captureMovieFrames(self):
        stim = visual.GratingStim(self.win, dkl=[0,0,1])
        stim.autoDraw = True
        for frameN in range(3):
            stim.phase += 0.3
            self.win.flip()
            self.win.getMovieFrame()
        self.win.saveMovieFrames(os.path.join(self.temp_dir, 'junkFrames.png'))
        self.win.saveMovieFrames(os.path.join(self.temp_dir, 'junkFrames.gif'))
        region = self.win._getRegionOfFrame()

    def test_multiFlip(self):
        self.win.recordFrameIntervals = False #does a reset
        self.win.recordFrameIntervals = True
        self.win.multiFlip(3)
        self.win.multiFlip(3,clearBuffer=False)
        self.win.saveFrameIntervals(os.path.join(self.temp_dir, 'junkFrameInts'))
        fps = self.win.fps()

    def test_callonFlip(self):
        def assertThisIs2(val):
            assert val==2
        self.win.callOnFlip(assertThisIs2, 2)
        self.win.flip()

class _baseVisualTest(object):
    #this class allows others to be created that inherit all the tests for
    #a different window config
    @classmethod
    def setup_class(self):#run once for each test class (window)
        self.win=None
        self.contextName
        raise NotImplementedError

    @classmethod
    def teardown_class(self):#run once for each test class (window)
        self.win.close()#shutil.rmtree(self.temp_dir)

    def setup(self):#this is run for each test individually
        #make sure we start with a clean window
        self.win.flip()

    def test_auto_draw(self):
        win = self.win
        stims=[]
        stims.append(visual.PatchStim(win))
        stims.append(visual.ShapeStim(win))
        stims.append(visual.TextStim(win))
        for stim in stims:
            assert stim.status==constants.NOT_STARTED
            stim.autoDraw = True
            assert stim.status==constants.STARTED
            stim.autoDraw = False
            assert stim.status==constants.FINISHED
            assert stim.status==constants.STOPPED
            "{}".format(stim) #check that str(xxx) is working

    def test_imageAndGauss(self):
        win = self.win
        fileName = os.path.join(utils.TESTS_DATA_PATH, 'testimage.jpg')
        #use image stim
        size = numpy.array([2.0,2.0])*self.scaleFactor
        image = visual.ImageStim(win, image=fileName, mask='gauss',
                                 size=size, flipHoriz=True, flipVert=True)
        image.draw()
        utils.compareScreenshot('imageAndGauss_%s.png' %(self.contextName), win)
        win.flip()

    def test_gratingImageAndGauss(self):
        win = self.win
        size = numpy.array([2.0,2.0])*self.scaleFactor
        #generate identical image as test_imageAndGauss but using GratingStim
        fileName = os.path.join(utils.TESTS_DATA_PATH, 'testimage.jpg')
        if win.units in ['norm','height']:
            sf = -1.0
        else:
            sf = -1.0 / size  # this will do the flipping and get exactly one cycle
        image = visual.GratingStim(win, tex=fileName, size=size, sf=sf, mask='gauss')
        image.draw()
        utils.compareScreenshot('imageAndGauss_%s.png' %(self.contextName), win)
        win.flip()

    def test_envelopeGratingAndRaisedCos(self):
        win = self.win
        size = numpy.array([2.0,2.0])*self.scaleFactor
        if win.units in ['norm','height']:
            sf = 5
        else:
            sf = 5.0/size #this will do the flipping and get exactly one cycle
        if win._haveShaders==True:  # can't draw envelope gratings without shaders so skip this test
            image = visual.EnvelopeGrating(win, carrier='sin', envelope='sin',
                                           size=size, sf=sf, mask='raisedCos',
                                           ori=-45, envsf=sf / 2, envori=45,
                                           envphase=90, moddepth=0.5,
                                           contrast=0.5)
            image.draw()
            utils.compareScreenshot('envelopeandrcos_%s.png' %(self.contextName), win)
            win.flip()
            "{}".format(image)
            
    def test_envelopeGratingPowerAndRaisedCos(self):
        win = self.win
        size = numpy.array([2.0,2.0])*self.scaleFactor
        if win.units in ['norm','height']:
            sf = 5
        else:
            sf = 5.0/size #this will do the flipping and get exactly one cycle
        if win._haveShaders==True:  # can't draw envelope gratings without shaders so skip this test
            image = visual.EnvelopeGrating(win, carrier='sin', envelope='sin',
                                           size=size, sf=sf, mask='raisedCos',
                                           ori=-45, envsf=sf / 2, envori=45,
                                           envphase=90, moddepth=0.5, power=0.5,
                                           contrast=0.5)
            image.draw()
            utils.compareScreenshot('envelopepowerandrcos_%s.png' %(self.contextName), win)
            win.flip()
            "{}".format(image)

    def test_NoiseStim_defaults(self):
        noiseTypes = ['binary', 'uniform', 'normal', 'white', 'filtered']

        for noiseType in noiseTypes:
            stim = visual.NoiseStim(win=self.win,
                                    noiseType=noiseType,
                                    size=(32, 32),
                                    units='pix')
            stim.updateNoise()
            stim.draw()

    def test_NoiseStim_defaults_image(self):
        noiseType = 'image'

        # noiseImage kwarg missing.
        with pytest.raises(ValueError):
            visual.NoiseStim(win=self.win,
                             noiseType=noiseType,
                             size=(32, 32),
                             units='pix')

    def test_noiseAndRaisedCos(self):
        numpy.random.seed(1)
        win = self.win
        size = numpy.array([2.0,2.0])*self.scaleFactor
        tres=128
        elementsize=4
        sf=None
        fileName = os.path.join(utils.TESTS_DATA_PATH, 'testimagegray.jpg')
        if win.units in ['pix']:
            ntype='Binary'
            size = numpy.array([128,128])
        elif win.units in ['degFlatPos']:
            ntype='Gabor'
            sf=0.125
        elif win.units in ['degFlat']:
            ntype='Isotropic'
            sf=0.125
        elif win.units in ['deg']:
            ntype='Filtered'
            sf=0.125
        elif win.units in ['cm']:
            ntype='Image'
            sf=0.25
        else:
            if self.contextName=='stencil':
                ntype='White'
            elif self.contextName=='height':
                ntype='Uniform'
            else:
                ntype='Normal'
            elementsize=1.0/8.0
        image  = visual.NoiseStim(win=win, name='noise',units=win.units, 
            noiseImage=fileName, mask='raisedCos',
            ori=0, pos=(0, 0), size=size, sf=sf, phase=0,
            color=[1,1,1], colorSpace='rgb', opacity=1, blendmode='avg', contrast=0.5,
            texRes=tres,
            noiseType=ntype, noiseElementSize=elementsize, noiseBaseSf=32.0/size[0],
            noiseBW=0.5, noiseBWO=7, noiseFractalPower=-1,noiseFilterLower=4.0/size[0], noiseFilterUpper=16.0/size[0], noiseFilterOrder=1, noiseClip=4.0, interpolate=False, depth=-1.0)
        image.draw()
        utils.compareScreenshot('noiseAndRcos_%s.png' %(self.contextName), win)
        win.flip()
        str(image)
        
    def test_noiseFiltersAndRaisedCos(self):
        numpy.random.seed(1)
        win = self.win
        size = numpy.array([2.0,2.0])*self.scaleFactor
        tres=128
        elementsize=4
        sf=None
        ntype='Binary'
        comp='Amplitude'
        fileName = os.path.join(utils.TESTS_DATA_PATH, 'testimagegray.jpg')
        if win.units in ['pix']:
            ftype='Butterworth'
            size = numpy.array([128,128])
        elif win.units in ['degFlatPos']:
            ftype='Gabor'
            sf=0.125
            elementsize=1
        elif win.units in ['degFlat']:
            ftype='Isotropic'
            sf=0.125
            elementsize=1
        elif win.units in ['deg']:
            ntype='Image'
            ftype='Butterworth'
            sf=0.125
        elif win.units in ['cm']:
            ntype='Image'
            ftype='Butterworth'
            comp='Phase'
            sf=0.25
        else:
            if self.contextName=='stencil':
                ntype='White'
                ftype='Butterworth'
            elif self.contextName=='height':
                ntype='Uniform'
                ftype='Butterworth'
            else:
                ntype='Normal'
                ftype='Butterworth'
            elementsize=1.0/8.0
        image  = visual.NoiseStim(win=win, name='noise',units=win.units, 
            noiseImage=fileName, mask='raisedCos',
            ori=0, pos=(0, 0), size=size, sf=sf, phase=0,
            color=[1,1,1], colorSpace='rgb', opacity=1, blendmode='avg', contrast=0.5,
            texRes=tres,
            noiseType=ntype, noiseElementSize=elementsize, noiseBaseSf=32.0/size[0],
            noiseBW=0.5, noiseBWO=7, noiseFractalPower=-1,noiseFilterLower=4.0/size[0], 
            noiseFilterUpper=16.0/size[0], noiseFilterOrder=1, noiseOri=45.0,
            noiseClip=4.0, imageComponent=comp, filter=ftype, interpolate=False, depth=-1.0)
 
        image.draw()
        utils.compareScreenshot('noiseFiltersAndRcos_%s.png' %(self.contextName), win)
        win.flip()
        str(image)

    def test_envelopeBeatAndRaisedCos(self):
        win = self.win
        size = numpy.array([2.0,2.0])*self.scaleFactor
        if win.units in ['norm','height']:
            sf = 5
        else:
            sf = 5.0/size #this will do the flipping and get exactly one cycle
        if win._haveShaders==True:  # can't draw envelope gratings without shaders so skip this test
            image = visual.EnvelopeGrating(win, carrier='sin', envelope='sin',
                                           size=size, sf=sf, mask='raisedCos',
                                           ori=-45, envsf=sf / 2, envori=45,
                                           envphase=90, beat=True, moddepth=0.5,
                                           contrast=0.5)
            image.draw()
            utils.compareScreenshot('beatandrcos_%s.png' %(self.contextName), win)
            win.flip()
            "{}".format(image)

    def test_numpyFilterMask(self):
        """if the mask is passed in as a numpy array it goes through a different
        set of rules when turned into a texture. But the outcome should be as above
        """
        win = self.win
        from psychopy.visual import filters
        gaussMask = filters.makeMask(128, 'gauss')
        size = numpy.array([2.0,2.0])*self.scaleFactor
        fileName = os.path.join(utils.TESTS_DATA_PATH, 'testimage.jpg')
        image = visual.ImageStim(win, image=fileName, mask=gaussMask,
                                 size=size, flipHoriz=True, flipVert=True)
        image.draw()
        utils.compareScreenshot('imageAndGauss_%s.png' %(self.contextName), win)
        win.flip()

    def test_greyscaleImage(self):
        win = self.win
        fileName = os.path.join(utils.TESTS_DATA_PATH, 'greyscale.jpg')
        imageStim = visual.ImageStim(win, fileName)
        imageStim.draw()
        utils.compareScreenshot('greyscale_%s.png' %(self.contextName), win)
        "{}".format(imageStim) #check that str(xxx) is working
        win.flip()
        imageStim.color = [0.1,0.1,0.1]
        imageStim.draw()
        utils.compareScreenshot('greyscaleLowContr_%s.png' %(self.contextName), win)
        win.flip()
        imageStim.color = 1
        imageStim.contrast = 0.1#should have identical effect to color=0.1
        imageStim.draw()
        utils.compareScreenshot('greyscaleLowContr_%s.png' %(self.contextName), win)
        win.flip()
        imageStim.contrast = 1.0
        fileName = os.path.join(utils.TESTS_DATA_PATH, 'greyscale2.png')
        imageStim.image = fileName
        imageStim.size *= 3
        imageStim.draw()
        utils.compareScreenshot('greyscale2_%s.png' %(self.contextName), win)
        win.flip()

    def test_numpyTexture(self):
        win = self.win
        grating = filters.makeGrating(res=64, ori=20.0,
                                     cycles=3.0, phase=0.5,
                                     gratType='sqr', contr=1.0)
        imageStim = visual.ImageStim(win, image=grating,
                                     size = 3*self.scaleFactor,
                                     interpolate=True)
        imageStim.draw()

        utils.compareScreenshot('numpyImage_%s.png' %(self.contextName), win)
        "{}".format(imageStim) #check that str(xxx) is working
        win.flip()
        #set lowcontrast using color
        imageStim.color = [0.1,0.1,0.1]
        imageStim.draw()
        utils.compareScreenshot('numpyLowContr_%s.png' %(self.contextName), win)
        win.flip()
        #now try low contrast using contr
        imageStim.color = 1
        imageStim.contrast = 0.1#should have identical effect to color=0.1
        imageStim.draw()
        utils.compareScreenshot('numpyLowContr_%s.png' %(self.contextName), win)
        win.flip()

    def test_hexColors(self):
        win = self.win
        circle = visual.Circle(win, fillColor='#0000FF',
                               lineColor=None,
                               size=2* self.scaleFactor)
        circle.draw()
        grat = visual.GratingStim(win, ori=20, color='#00AAFF',
            pos=[0.6 * self.scaleFactor, -0.6 * self.scaleFactor],
            sf=3.0 / self.scaleFactor, size=2 * self.scaleFactor,
            interpolate=True)
        grat.draw()
        utils.compareScreenshot('circleHex_%s.png' %(self.contextName), win)
        win.flip()


    def test_gabor(self):
        win = self.win
        #using init
        gabor = visual.PatchStim(win, mask='gauss', ori=-45,
            pos=[0.6 * self.scaleFactor, -0.6 * self.scaleFactor],
            sf=2.0 / self.scaleFactor, size=2 * self.scaleFactor,
            interpolate=True)
        gabor.draw()
        utils.compareScreenshot('gabor1_%s.png' %(self.contextName), win)
        win.flip()#AFTER compare screenshot

        #using .set()
        gabor.ori = 45
        gabor.size -= 0.2 * self.scaleFactor
        gabor.setColor([45, 30, 0.3], colorSpace='dkl')
        gabor.sf += 0.2 / self.scaleFactor
        gabor.pos += [-0.5*self.scaleFactor, 0.5*self.scaleFactor]
        gabor.contrast = 0.8
        gabor.opacity = 0.8
        gabor.draw()
        utils.compareScreenshot('gabor2_%s.png' %(self.contextName), win)
        win.flip()
        "{}".format(gabor) #check that str(xxx) is working

    @pytest.mark.bufferimage
    def test_bufferImage(self):
        """BufferImage inherits from ImageStim, so test .ori. .pos etc there not here
        """
        win = self.win
        gabor = visual.PatchStim(win, mask='gauss', ori=-45,
            pos=[0.6*self.scaleFactor, -0.6*self.scaleFactor],
            sf=2.0/self.scaleFactor, size=2*self.scaleFactor,
            interpolate=True)

        bufferImgStim = visual.BufferImageStim(self.win, stim=[gabor],
            interpolate=True)
        bufferImgStim.draw()
        utils.compareScreenshot('bufferimg_gabor_%s.png' %(self.contextName), win, crit=8)
        win.flip()

    #def testMaskMatrix(self):
    #    #aims to draw the exact same stimulus as in testGabor, but using filters
    #    win=self.win
    #    contextName=self.contextName
    #    #create gabor using filters
    #    size=2*self.scaleFactor#to match Gabor1 above
    #    if win.units in ['norm','height']:
    #        sf=1.0/size
    #    else:
    #        sf=2.0/self.scaleFactor#to match Gabor1 above
    #    cycles=size*sf
    #    grating = filters.makeGrating(256, ori=135, cycles=cycles)
    #    gabor = filters.maskMatrix(grating, shape='gauss')
    #    stim = visual.PatchStim(win, tex=gabor,
    #        pos=[0.6*self.scaleFactor, -0.6*self.scaleFactor],
    #        sf=1.0/size, size=size,
    #        interpolate=True)
    #    stim.draw()
    #    utils.compareScreenshot('gabor1_%s.png' %(contextName), win)

    def test_text(self):
        win = self.win
        #set font
        fontFile = os.path.join(prefs.paths['resources'], 'DejaVuSerif.ttf')
        #using init
        stim = visual.TextStim(win,text=u'\u03A8a', color=[0.5, 1.0, 1.0], ori=15,
            height=0.8*self.scaleFactor, pos=[0,0], font='DejaVu Serif',
            fontFiles=[fontFile])
        stim.draw()
        if self.win.winType != 'pygame':
            #compare with a LIBERAL criterion (fonts do differ)
            utils.compareScreenshot('text1_%s.png' %(self.contextName), win, crit=20)
        win.flip()#AFTER compare screenshot
        #using set
        stim.text = 'y'
        if sys.platform=='win32':
            stim.font = 'Courier New'
        else:
            stim.font = 'Courier'
        stim.ori = -30.5
        stim.height = 1.0 * self.scaleFactor
        stim.setColor([0.1, -1, 0.8], colorSpace='rgb')
        stim.pos += [-0.5, 0.5]
        stim.contrast = 0.8
        stim.opacity = 0.8
        stim.draw()
        "{}".format(stim) #check that str(xxx) is working
        if self.win.winType != 'pygame':
            #compare with a LIBERAL criterion (fonts do differ)
            utils.compareScreenshot('text2_%s.png' %self.contextName,
                                    win, crit=20)

    def test_text_with_add(self):
        # pyglet text will reset the blendMode to 'avg' so check that we are
        # getting back to 'add' if we want it
        win = self.win
        text = visual.TextStim(win, pos=[0, 0.9])
        grat1 = visual.GratingStim(win, size=2*self.scaleFactor,
                                   opacity=0.5,
                                   pos=[0.3,0.0], ori=45, sf=2*self.scaleFactor)
        grat2 = visual.GratingStim(win, size=2 * self.scaleFactor,
                                   opacity=0.5,
                                   pos=[-0.3, 0.0], ori=-45,
                                   sf=2*self.scaleFactor)

        text.draw()
        grat1.draw()
        grat2.draw()
        utils.skip_under_travis()
        if self.win.winType != 'pygame':
            utils.compareScreenshot('blend_add_%s.png' %self.contextName,
                                    win, crit=20)

    def test_mov(self):
        win = self.win
        if self.win.winType == 'pygame':
            pytest.skip("movies only available for pyglet backend")
        elif _travisTesting and not _anacondaTesting:
            pytest.skip("Travis with system Python doesn't seem to have a "
                        "working ffmpeg")
        win.flip()
        #construct full path to the movie file
        fileName = os.path.join(utils.TESTS_DATA_PATH, 'testMovie.mp4')
        #check if present
        if not os.path.isfile(fileName):
            raise IOError('Could not find movie file: %s'
                          % os.path.abspath(fileName))
        #then do actual drawing
        pos = [0.6*self.scaleFactor, -0.6*self.scaleFactor]
        mov = visual.MovieStim3(win, fileName, pos=pos, noAudio=True)
        mov.setFlipVert(True)
        mov.setFlipHoriz(True)
        for frameN in range(10):
            mov.draw()
            if frameN==0:
                utils.compareScreenshot('movFrame1_%s.png' %self.contextName,
                                        win, crit=10)
            win.flip()
        "{}".format(mov) #check that str(xxx) is working

    def test_rect(self):
        win = self.win
        rect = visual.Rect(win)
        rect.draw()
        rect.lineColor = 'blue'
        rect.pos = [1, 1]
        rect.ori = 30
        rect.fillColor = 'pink'
        rect.draw()
        "{}".format(rect) #check that str(xxx) is working
        rect.width = 1
        rect.height = 1

    def test_circle(self):
        win = self.win
        circle = visual.Circle(win)
        circle.fillColor = 'red'
        circle.draw()
        circle.lineColor = 'blue'
        circle.fillColor = None
        circle.pos = [0.5, -0.5]
        circle.ori = 30
        circle.draw()
        "{}".format(circle) #check that str(xxx) is working

    def test_line(self):
        win = self.win
        line = visual.Line(win)
        line.start = (0, 0)
        line.end = (0.1, 0.1)
        line.draw()
        win.flip()
        "{}".format(line)  # check that str(xxx) is working

    def test_Polygon(self):
        win = self.win
        cols = ['red','green','purple','orange','blue']
        for n, col in enumerate(cols):
            poly = visual.Polygon(win, edges=n + 5, lineColor=col)
            poly.draw()
        win.flip()
        "{}".format(poly) #check that str(xxx) is working
        poly.edges = 3
        poly.radius = 1

    @pytest.mark.shape2
    def test_shape(self):
        win = self.win
        arrow = [(-0.4,0.05), (-0.4,-0.05), (-.2,-0.05), (-.2,-0.1), (0,0), (-.2,0.1), (-.2,0.05)]
        shape = visual.ShapeStim(win, lineColor='white', lineWidth=1.0,
            fillColor='red', vertices=arrow, pos=[0, 0],
            ori=0.0, opacity=1.0, depth=0, interpolate=True)
        shape.draw()
        #NB shape rendering can differ a little, depending on aliasing
        utils.compareScreenshot('shape2_1_%s.png' %(self.contextName), win, crit=12.5)
        win.flip()

        # Using .set()
        shape.contrast = 0.8
        shape.opacity = 0.8
        shape.ori = 90
        shape.draw()
        assert 'Shape' in "{}".format(shape)  # check that str(xxx) is working
        utils.compareScreenshot('shape2_2_%s.png' %(self.contextName), win, crit=12.5)

    def test_radial(self):
        win = self.win
        #using init
        wedge = visual.RadialStim(win, tex='sqrXsqr', color=1,size=2*self.scaleFactor,
            visibleWedge=[0, 45], radialCycles=2, angularCycles=2, interpolate=False)
        wedge.draw()
        thresh = 15  # there are often a slight interpolation differences
        if win.winType != 'pygame':  # pygame definitely gets radialstim wrong!
            utils.compareScreenshot('wedge1_%s.png' %(self.contextName),
                                    win, crit=thresh)
        win.flip()#AFTER compare screenshot

        #using .set()
        wedge.mask = 'gauss'
        wedge.size = 3 * self.scaleFactor
        wedge.angularCycles = 3
        wedge.radialCycles = 3
        wedge.ori = 180
        wedge.contrast = 0.8
        wedge.opacity = 0.8
        wedge.radialPhase += 0.1
        wedge.angularPhase = 0.1
        wedge.draw()
        "{}".format(wedge) #check that str(xxx) is working
        if win.winType != 'pygame' and (sys.version_info.major, sys.version_info.minor) != (3, 6):  # pygame and 3.6 definitely get radialstim wrong!
            utils.compareScreenshot('wedge2_%s.png' %(self.contextName),
                                    win, crit=thresh)
        else:
            pytest.skip("Pygame fails to render RadialStim properly :-/")

    def test_simpleimage(self):
        win = self.win
        if win.useRetina:
            pytest.skip("Rendering pixel-for-pixel is not identical on retina")
        fileName = os.path.join(utils.TESTS_DATA_PATH, 'testimage.jpg')
        if not os.path.isfile(fileName):
            raise IOError('Could not find image file: %s' % os.path.abspath(fileName))
        image = visual.SimpleImageStim(win, image=fileName, flipHoriz=True, flipVert=True)
        "{}".format(image) #check that str(xxx) is working
        image.draw()
        utils.compareScreenshot('simpleimage1_%s.png' %(self.contextName), win, crit=5.0) # Should be exact replication

    def test_dotsUnits(self):
        #to test this create a small dense circle of dots and check the circle
        #has correct dimensions
        fieldSize = numpy.array([1.0,1.0])*self.scaleFactor
        pos = numpy.array([0.5,0])*fieldSize
        dots = visual.DotStim(self.win, color=[-1.0,0.0,0.5], dotSize=5,
                              nDots=1000, fieldShape='circle', fieldPos=pos)
        dots.draw()
        utils.compareScreenshot('dots_%s.png' %(self.contextName), self.win, crit=20)
        self.win.flip()

    def test_dots(self):
        #NB we can't use screenshots here - just check that no errors are raised
        win = self.win
        #using init
        dots =visual.DotStim(win, color=(1.0,1.0,1.0), dir=270,
            nDots=500, fieldShape='circle', fieldPos=(0.0,0.0),fieldSize=1*self.scaleFactor,
            dotLife=5, #number of frames for each dot to be drawn
            signalDots='same', #are the signal and noise dots 'different' or 'same' popns (see Scase et al)
            noiseDots='direction', #do the noise dots follow random- 'walk', 'direction', or 'position'
            speed=0.01*self.scaleFactor, coherence=0.9)
        dots.draw()
        win.flip()
        "{}".format(dots) #check that str(xxx) is working

        #using .set() and check the underlying variable changed
        prevDirs = copy.copy(dots._dotsDir)
        prevSignals = copy.copy(dots._signalDots)
        prevVerticesPix = copy.copy(dots.verticesPix)
        dots.dir = 20
        dots.coherence = 0.5
        dots.fieldPos = [-0.5, 0.5]
        dots.speed = 0.1 * self.scaleFactor
        dots.opacity = 0.8
        dots.contrast = 0.8
        dots.draw()
        #check that things changed
        assert (prevDirs-dots._dotsDir).sum()!=0, \
            "dots._dotsDir failed to change after dots.setDir()"
        assert prevSignals.sum()!=dots._signalDots.sum(), \
            "dots._signalDots failed to change after dots.setCoherence()"
        assert not numpy.alltrue(prevVerticesPix==dots.verticesPix), \
            "dots.verticesPix failed to change after dots.setPos()"

    def test_element_array(self):
        win = self.win
        if not win._haveShaders:
            pytest.skip("ElementArray requires shaders, which aren't available")
        #using init
        thetas = numpy.arange(0,360,10)
        N=len(thetas)

        radii = numpy.linspace(0,1.0,N)*self.scaleFactor
        x, y = pol2cart(theta=thetas, radius=radii)
        xys = numpy.array([x,y]).transpose()
        spiral = visual.ElementArrayStim(
                win, opacities = 0, nElements=N, sizes=0.5*self.scaleFactor,
                sfs=1.0, xys=xys, oris=-thetas)
        spiral.draw()
        #check that the update function is working by changing vals after first draw() call
        spiral.opacities = 1.0
        spiral.sfs = 3.0
        spiral.draw()
        "{}".format(spiral) #check that str(xxx) is working
        win.flip()
        spiral.draw()
        utils.compareScreenshot('elarray1_%s.png' %(self.contextName), win)
        win.flip()

    def test_aperture(self):
        win = self.win
        if not win.allowStencil:
            pytest.skip("Don't run aperture test when no stencil is available")
        grating = visual.GratingStim(
                win, mask='gauss',sf=8.0, size=2,color='FireBrick',
                units='norm')
        aperture = visual.Aperture(win, size=1*self.scaleFactor,
                                   pos=[0.8*self.scaleFactor,0])
        aperture.enabled = False
        grating.draw()
        aperture.enabled = True
        "{}".format(aperture) #check that str(xxx) is working
        grating.ori = 90
        grating.color = 'black'
        grating.draw()
        utils.compareScreenshot('aperture1_%s.png' %(self.contextName), win)
        #aperture should automatically disable on exit
        for shape, nVert, pos in [(None, 120, (0,0)), ('circle', 17, (.2, -.7)),
                                  ('square', 4, (-.5,-.5)), ('triangle', 3, (1,1))]:
            aperture = visual.Aperture(win, pos=pos, shape=shape, nVert=nVert)
            assert len(aperture.vertices) == nVert  # true for BaseShapeStim; expect (nVert-2)*3 if tesselated
            assert aperture.contains(pos)

    def test_aperture_image(self):
        win = self.win
        fileName = os.path.join(utils.TESTS_DATA_PATH, 'testwedges.png')
        if not win.allowStencil:
            pytest.skip("Don't run aperture test when no stencil is available")
        grating = visual.GratingStim(win, mask='gauss',sf=8.0, size=2,
                                     color='FireBrick', units='norm')
        aperture = visual.Aperture(win, size=1*self.scaleFactor,
                                   pos=[0.8*self.scaleFactor,0], shape=fileName)
        aperture.enabled = False
        grating.draw()
        aperture.enabled = True
        "{}".format(aperture) #check that str(xxx) is working
        grating.ori = 90
        grating.color = 'black'
        grating.draw()
        utils.compareScreenshot('aperture2_%s.png' %(self.contextName),
                                win, crit=30)
        #aperture should automatically disable on exit

    def test_rating_scale(self):
        if self.win.winType=='pygame':
            pytest.skip("RatingScale not available on pygame")
        # try to avoid text; avoid default / 'triangle' because it does not display on win XP
        win = self.win
        win.flip()
        rs = visual.RatingScale(win, low=0, high=1, precision=100, size=3, pos=(0,-.4),
                        labels=[' ', ' '], scale=' ',
                        marker='glow', markerStart=0.7, markerColor='darkBlue', autoLog=False)
        "{}".format(rs) #check that str(xxx) is working
        rs.draw()
        utils.compareScreenshot('ratingscale1_%s.png' %(self.contextName), win, crit=40.0)
        win.flip()#AFTER compare screenshot

    def test_refresh_rate(self):
        if self.win.winType=='pygame':
            pytest.skip("getMsPerFrame seems to crash the testing of pygame")
        #make sure that we're successfully syncing to the frame rate
        msPFavg, msPFstd, msPFmed = visual.getMsPerFrame(self.win,nFrames=60, showVisual=True)
        utils.skip_under_travis()             # skip late so we smoke test the code
        assert (1000/150.0) < msPFavg < (1000/40.0), \
            "Your frame period is %.1fms which suggests you aren't syncing to the frame" %msPFavg


#create different subclasses for each context/backend
class TestPygletNorm(_baseVisualTest):
    @classmethod
    def setup_class(self):
        self.win = visual.Window([128,128], winType='pyglet', pos=[50,50],
                                 allowStencil=True, autoLog=False)
        self.contextName='norm'
        self.scaleFactor=1#applied to size/pos values

class TestPygletHexColor(_baseVisualTest):
    @classmethod
    def setup_class(self):
        self.win = visual.Window([128,128], winType='pyglet', pos=[50,50],
                                 color="#FF0099",
                                 allowStencil=True, autoLog=False)
        self.contextName='normHexbackground'
        self.scaleFactor=1#applied to size/pos values

if not _travisTesting:
    class TestPygletBlendAdd(_baseVisualTest):
        @classmethod
        def setup_class(self):
            self.win = visual.Window([128,128], winType='pyglet', pos=[50,50],
                                     blendMode='add', useFBO=True)
            self.contextName='normAddBlend'
            self.scaleFactor=1#applied to size/pos values


class TestPygletNormFBO(_baseVisualTest):
    @classmethod
    def setup_class(self):
        self.win = visual.Window([128,128], winType='pyglet', pos=[50,50],
                                 allowStencil=True, autoLog=False, useFBO=True)
        self.contextName='norm'
        self.scaleFactor=1#applied to size/pos values


class TestPygletHeight(_baseVisualTest):
    @classmethod
    def setup_class(self):
        self.win = visual.Window([128,64], winType='pyglet', pos=[50,50],
                                 allowStencil=False, autoLog=False)
        self.contextName='height'
        self.scaleFactor=1#applied to size/pos values


class TestPygletNormNoShaders(_baseVisualTest):
    @classmethod
    def setup_class(self):
        self.win = visual.Window([128,128], monitor='testMonitor',
                                 winType='pyglet', pos=[50,50],
                                 allowStencil=True, autoLog=False)
        self.win._haveShaders=False
        self.contextName='normNoShade'
        self.scaleFactor=1#applied to size/pos values


class TestPygletNormStencil(_baseVisualTest):
    @classmethod
    def setup_class(self):
        self.win = visual.Window([128,128], monitor='testMonitor',
                                 winType='pyglet', pos=[50,50],
                                 allowStencil=True, autoLog=False)
        self.contextName='stencil'
        self.scaleFactor=1#applied to size/pos values


class TestPygletPix(_baseVisualTest):
    @classmethod
    def setup_class(self):
        mon = monitors.Monitor('testMonitor')
        mon.setDistance(57)
        mon.setWidth(40.0)
        mon.setSizePix([1024,768])
        self.win = visual.Window([128,128], monitor=mon, winType='pyglet', pos=[50,50], allowStencil=True,
            units='pix', autoLog=False)
        self.contextName='pix'
        self.scaleFactor=60#applied to size/pos values


class TestPygletCm(_baseVisualTest):
    @classmethod
    def setup_class(self):
        mon = monitors.Monitor('testMonitor')
        mon.setDistance(57.0)
        mon.setWidth(40.0)
        mon.setSizePix([1024,768])
        self.win = visual.Window([128,128], monitor=mon, winType='pyglet', pos=[50,50], allowStencil=False,
            units='cm', autoLog=False)
        self.contextName='cm'
        self.scaleFactor=2#applied to size/pos values


class TestPygletDeg(_baseVisualTest):
    @classmethod
    def setup_class(self):
        mon = monitors.Monitor('testMonitor')
        mon.setDistance(57.0)
        mon.setWidth(40.0)
        mon.setSizePix([1024,768])
        self.win = visual.Window([128,128], monitor=mon, winType='pyglet', pos=[50,50], allowStencil=True,
            units='deg', autoLog=False)
        self.contextName='deg'
        self.scaleFactor=2#applied to size/pos values


class TestPygletDegFlat(_baseVisualTest):
    @classmethod
    def setup_class(self):
        mon = monitors.Monitor('testMonitor')
        mon.setDistance(10.0) #exagerate the effect of flatness by setting the monitor close
        mon.setWidth(40.0)
        mon.setSizePix([1024,768])
        self.win = visual.Window([128,128], monitor=mon, winType='pyglet', pos=[50,50], allowStencil=True,
            units='degFlat', autoLog=False)
        self.contextName='degFlat'
        self.scaleFactor=4#applied to size/pos values


class TestPygletDegFlatPos(_baseVisualTest):
    @classmethod
    def setup_class(self):
        mon = monitors.Monitor('testMonitor')
        mon.setDistance(10.0) #exagerate the effect of flatness by setting the monitor close
        mon.setWidth(40.0)
        mon.setSizePix([1024,768])
        self.win = visual.Window([128,128], monitor=mon, winType='pyglet', pos=[50,50], allowStencil=True,
            units='degFlatPos', autoLog=False)
        self.contextName='degFlatPos'
        self.scaleFactor=4#applied to size/pos values

@pytest.mark.needs_pygame
class TestPygameNorm(_baseVisualTest):
   @classmethod
   def setup_class(self):
       self.win = visual.Window([128,128], winType='pygame', allowStencil=True, autoLog=False)
       self.contextName='norm'
       self.scaleFactor=1#applied to size/pos values
#class TestPygamePix(_baseVisualTest):
#    @classmethod
#    def setup_class(self):
#        mon = monitors.Monitor('testMonitor')
#        mon.setDistance(57.0)
#        mon.setWidth(40.0)
#        mon.setSizePix([1024,768])
#        self.win = visual.Window([128,128], monitor=mon, winType='pygame', allowStencil=True,
#            units='pix', autoLog=False)
#        self.contextName='pix'
#        self.scaleFactor=60#applied to size/pos values
#class TestPygameCm(_baseVisualTest):
#    @classmethod
#    def setup_class(self):
#        mon = monitors.Monitor('testMonitor')
#        mon.setDistance(57.0)
#        mon.setWidth(40.0)
#        mon.setSizePix([1024,768])
#        self.win = visual.Window([128,128], monitor=mon, winType='pygame', allowStencil=False,
#            units='cm')
#        self.contextName='cm'
#        self.scaleFactor=2#applied to size/pos values
#class TestPygameDeg(_baseVisualTest):
#    @classmethod
#    def setup_class(self):
#        mon = monitors.Monitor('testMonitor')
#        mon.setDistance(57.0)
#        mon.setWidth(40.0)
#        mon.setSizePix([1024,768])
#        self.win = visual.Window([128,128], monitor=mon, winType='pygame', allowStencil=True,
#            units='deg')
#        self.contextName='deg'
#        self.scaleFactor=2#applied to size/pos values
#


if __name__ == '__main__':
    cls = TestPygletDegFlatPos()
    cls.setup_class()
    cls.test_radial()
    cls.teardown_class()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Part of the PsychoPy library
# Copyright (C) 2002-2018 Jonathan Peirce (C) 2019-2021 Open Science Tools Ltd.
# Distributed under the terms of the GNU General Public License (GPL).

"""Error dialog for showing unhandled exceptions that occur within the PsychoPy
app."""

import wx
import traceback
import psychopy.preferences
import sys
from psychopy.localization import _translate


_error_dlg = None  # keep error dialogs from stacking


class ErrorMsgDialog(wx.Dialog):
    """Class for creating an error report dialog. Should never be created
    directly.
    """
    def __init__(self, parent, traceback=''):
        wx.Dialog.__init__(self, parent, id=wx.ID_ANY,
                           title=_translate(u"PsychoPy3 Error"),
                           pos=wx.DefaultPosition, size=wx.Size(750, -1),
                           style=wx.DEFAULT_DIALOG_STYLE | wx.RESIZE_BORDER)

        self.details = traceback

        # message to show at the top of the error box, needs translation
        msg = _translate(u"PsychoPy encountered an unhandled internal error! " \
                u"Please send the report under \"Details\" to the " \
                u"developers with a description of what you were doing " \
                u"with the software when the error occurred.")

        self.SetSizeHints(wx.DefaultSize, wx.DefaultSize)
        szErrorMsg = wx.BoxSizer(wx.VERTICAL)
        szHeader = wx.FlexGridSizer(0, 3, 0, 0)
        szHeader.AddGrowableCol(1)
        szHeader.SetFlexibleDirection(wx.BOTH)
        szHeader.SetNonFlexibleGrowMode(wx.FLEX_GROWMODE_SPECIFIED)
        self.imgErrorIcon = wx.StaticBitmap(
            self, wx.ID_ANY, wx.ArtProvider.GetBitmap(
                wx.ART_ERROR, wx.ART_MESSAGE_BOX),
            wx.DefaultPosition, wx.DefaultSize, 0)
        szHeader.Add(self.imgErrorIcon, 0, wx.ALL, 5)
        self.lblErrorMsg = wx.StaticText(
            self, wx.ID_ANY, msg, wx.DefaultPosition, wx.DefaultSize, 0)
        self.lblErrorMsg.Wrap(560)
        szHeader.Add(self.lblErrorMsg, 0, wx.ALL, 5)
        szHeaderButtons = wx.BoxSizer(wx.VERTICAL)
        self.cmdOK = wx.Button(
            self, wx.ID_OK, _translate(u"&OK"), wx.DefaultPosition, wx.DefaultSize, 0)
        szHeaderButtons.Add(self.cmdOK, 0, wx.LEFT | wx.EXPAND, 5)
        self.cmdExit = wx.Button(
            self, wx.ID_EXIT, _translate(u"E&xit PsychoPy"), wx.DefaultPosition,
            wx.DefaultSize, 0)
        szHeaderButtons.Add(self.cmdExit, 0, wx.TOP | wx.LEFT | wx.EXPAND, 5)
        szHeader.Add(szHeaderButtons, 0, wx.ALL | wx.EXPAND, 5)
        szErrorMsg.Add(szHeader, 0, wx.TOP | wx.LEFT | wx.RIGHT | wx.EXPAND, 5)

        self.pnlDetails = wx.CollapsiblePane(
            self, wx.ID_ANY, _translate(u"&Details"), wx.DefaultPosition, wx.DefaultSize,
            wx.CP_DEFAULT_STYLE)
        self.pnlDetails.Collapse(True)
        szDetailsPane = wx.BoxSizer(wx.VERTICAL)
        self.txtErrorOutput = wx.TextCtrl(
            self.pnlDetails.GetPane(), wx.ID_ANY, self.details,
            wx.DefaultPosition, wx.Size(640, 150),
            wx.TE_AUTO_URL | wx.TE_BESTWRAP | wx.TE_MULTILINE | wx.TE_READONLY |
            wx.TE_WORDWRAP)
        szDetailsPane.Add(self.txtErrorOutput, 1, wx.ALL | wx.EXPAND, 5)
        szTextButtons = wx.BoxSizer(wx.HORIZONTAL)
        self.cmdCopyError = wx.Button(
            self.pnlDetails.GetPane(), wx.ID_ANY, _translate(u"&Copy"), wx.DefaultPosition,
            wx.DefaultSize, 0)
        szTextButtons.Add(self.cmdCopyError, 0, wx.RIGHT, 5)
        self.cmdSaveError = wx.Button(
            self.pnlDetails.GetPane(), wx.ID_ANY, _translate(u"&Save"), wx.DefaultPosition,
            wx.DefaultSize, 0)
        szTextButtons.Add(self.cmdSaveError, 0)
        szDetailsPane.Add(szTextButtons, 0, wx.ALL | wx.ALIGN_RIGHT, 5)
        self.pnlDetails.Expand()
        self.pnlDetails.GetPane().SetSizer(szDetailsPane)
        self.pnlDetails.GetPane().Layout()
        szDetailsPane.Fit(self.pnlDetails.GetPane())
        szErrorMsg.Add(self.pnlDetails, 1, wx.ALL | wx.BOTTOM | wx.EXPAND, 5)

        self.SetSizer(szErrorMsg)
        self.Layout()
        self.Fit()

        self.Centre(wx.BOTH)

        # Connect Events
        self.cmdOK.Bind(wx.EVT_BUTTON, self.onOkay)
        self.cmdExit.Bind(wx.EVT_BUTTON, self.onExit)
        self.cmdCopyError.Bind(wx.EVT_BUTTON, self.onCopyDetails)
        self.cmdSaveError.Bind(wx.EVT_BUTTON, self.onSaveDetails)

        # ding!
        wx.Bell()

    def __del__(self):
        pass

    def onOkay(self, event):
        """Called when OK is clicked."""
        event.Skip()

    def onExit(self, event):
        """Called when the user requests to close PsychoPy. This can be called
        if the error if unrecoverable or if the errors are being constantly
        generated.

        Will try to close things safely, allowing the user to save files while
        suppressing further errors.

        """
        dlg = wx.MessageDialog(
            self,
            _translate("Are you sure you want to exit PsychoPy? Unsaved work may be lost "
            "(but we'll try to save opened files)."),
            _translate("Exit PsychoPy?"),
            wx.YES_NO | wx.NO_DEFAULT | wx.ICON_WARNING | wx.CENTRE)
        if dlg.ShowModal() == wx.ID_YES:
            wx.GetApp().quit()
            # wx.Exit()  # nuclear option
        else:
            dlg.Destroy()
            event.Skip()

    def onCopyDetails(self, event):
        """Copy the contents of the details text box to the clipboard. This is
        to allow the user to paste the traceback into an email, forum post,
        issue ticket, etc. to report the error to the developers. If there is a
        selection range, only that text will be copied.

        """
        # check if we have a selection
        start, end = self.txtErrorOutput.GetSelection()
        if start != end:
            txt = self.txtErrorOutput.GetStringSelection()
        else:
            txt = self.txtErrorOutput.GetValue()

        if wx.TheClipboard.Open():
            wx.TheClipboard.SetData(wx.TextDataObject(txt))
            wx.TheClipboard.Close()

        event.Skip()

    def onSaveDetails(self, event):
        """Dump the traceback data to a file. This can be used to save the error
        data so it can be reported to the developers at a later time. Brings up
        a file save dialog to select where to write the file.

        """
        with wx.FileDialog(
                self, _translate("Save error traceback"),
                wildcard="Text files (*.txt)|*.txt",
                defaultFile='psychopy_traceback.txt',
                style=wx.FD_SAVE | wx.FD_OVERWRITE_PROMPT) as fileDialog:

            if fileDialog.ShowModal() == wx.ID_CANCEL:
                return  # the user changed their mind

            # dump traceback to file
            pathname = fileDialog.GetPath()
            try:
                with open(pathname, 'w') as file:
                    file.write(self.txtErrorOutput.GetValue())
            except IOError:
                # error in an error ... ;)
                errdlg = wx.MessageDialog(
                    self,
                    _translate("Cannot save to file '%s'.") % pathname,
                    _translate("File save error"),
                    wx.OK_DEFAULT | wx.ICON_ERROR | wx.CENTRE)
                errdlg.ShowModal()
                errdlg.Destroy()

        event.Skip()


def isErrorDialogVisible():
    """Check if the error dialog is open. This can be used to prevent background
    routines from running while the user deals with an error.

    Returns
    -------
    bool
        Error dialog is currently active.

    """
    return _error_dlg is not None


def exceptionCallback(exc_type, exc_value, exc_traceback):
    """Hook when an unhandled exception is raised within the current application
    thread. Gets the exception message and creates an error dialog box.

    When this function is patched into `sys.excepthook`, all unhandled
    exceptions will result in a dialog being displayed.

    """
    if psychopy.preferences.prefs.app['errorDialog'] is False:
        # have the error go out to stdout if dialogs are disabled
        traceback.print_exception(
            exc_type, exc_value, exc_traceback, file=sys.stdout)
        return

    global _error_dlg
    if not isErrorDialogVisible():
        # format the traceback text
        tbText = ''.join(traceback.format_exception(
            exc_type, exc_value, exc_traceback))
        _error_dlg = ErrorMsgDialog(None, tbText)

        # show the dialog
        _error_dlg.ShowModal()
        _error_dlg.Destroy()
        _error_dlg = None



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Part of the PsychoPy library
# Copyright (C) 2002-2018 Jonathan Peirce (C) 2019-2021 Open Science Tools Ltd.
# Distributed under the terms of the GNU General Public License (GPL).

# Acknowledgements:
#    This code was mostly written by Jon Peirce.
#    CRS Ltd provided support as needed.
#    Shader code for mono++ and color++ modes was based on code in
#    Psychtoolbox (Kleiner) but does not actually use that code directly

"""Cambridge Research Systems makes devices to support particularly
vision research.
"""

from __future__ import absolute_import, print_function

from psychopy import logging
from .bits import BitsSharp, BitsPlusPlus, DisplayPlusPlus, DisplayPlusPlusTouch
from .colorcal import ColorCAL
# Monkey-patch our metadata into CRS class.
setattr(ColorCAL, "longName", "CRS ColorCAL")
setattr(ColorCAL, "driverFor", ["colorcal"])



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""A stimulus class for playing movies (mpeg, avi, etc...) in PsychoPy.
"""

# Part of the PsychoPy library
# Copyright (C) 2002-2018 Jonathan Peirce (C) 2019-2021 Open Science Tools Ltd.
# Distributed under the terms of the GNU General Public License (GPL).

from __future__ import absolute_import, division, print_function

from builtins import next
from builtins import str
import sys
import os

# Ensure setting pyglet.options['debug_gl'] to False is done prior to any
# other calls to pyglet or pyglet submodules, otherwise it may not get picked
# up by the pyglet GL engine and have no effect.
# Shaders will work but require OpenGL2.0 drivers AND PyOpenGL3.0+
import pyglet
pyglet.options['debug_gl'] = False
GL = pyglet.gl

# on windows try to load avbin now (other libs can interfere)
if sys.platform == 'win32':
    # make sure we also check in SysWOW64 if on 64-bit windows
    if 'C:\\Windows\\SysWOW64' not in os.environ['PATH']:
        os.environ['PATH'] += ';C:\\Windows\\SysWOW64'

    try:
        from pyglet.media import avbin
        haveAvbin = True
    except ImportError:
        # either avbin isn't installed or scipy.stats has been imported
        # (prevents avbin loading)
        haveAvbin = False
    except Exception as e:
        # WindowsError on some systems
        # AttributeError if using avbin5 from pyglet 1.2?
        haveAvbin = False


import psychopy  # so we can get the __path__
from psychopy import core, logging, event
import psychopy.event

# tools must only be imported *after* event or MovieStim breaks on win32
# (JWP has no idea why!)
from psychopy.tools.arraytools import val2array
from psychopy.tools.attributetools import logAttrib, setAttribute
from psychopy.visual.basevisual import BaseVisualStim, ContainerMixin
from psychopy.tools.filetools import pathToString

if sys.platform == 'win32' and not haveAvbin:
    logging.warning("avbin.dll failed to load. "
                    "Try importing psychopy.visual as the first library "
                    "(before anything that uses scipy) or use a different"
                    "movie backend (e.g. moviepy).")

import numpy
try:
    from pyglet import media
    havePygletMedia = True
except Exception:
    havePygletMedia = False

from psychopy.constants import FINISHED, NOT_STARTED, PAUSED, PLAYING, STOPPED


class MovieStim(BaseVisualStim, ContainerMixin):
    """A stimulus class for playing movies (mpeg, avi, etc...) in PsychoPy.

    **Example**::

        mov = visual.MovieStim(myWin, 'testMovie.mp4', flipVert=False)
        print(mov.duration)
        # give the original size of the movie in pixels:
        print(mov.format.width, mov.format.height)

        mov.draw()  # draw the current frame (automagically determined)

    See MovieStim.py for demo.
    """

    def __init__(self, win,
                 filename="",
                 units='pix',
                 size=None,
                 pos=(0.0, 0.0),
                 ori=0.0,
                 flipVert=False,
                 flipHoriz=False,
                 color=(1.0, 1.0, 1.0),
                 colorSpace='rgb',
                 opacity=1.0,
                 volume=1.0,
                 name=None,
                 loop=False,
                 autoLog=None,
                 depth=0.0,):
        """
        :Parameters:

            filename :
                a string giving the relative or absolute path to the movie.
                Can be any movie that AVbin can read (e.g. mpeg, DivX)
            flipVert : True or *False*
                If True then the movie will be top-bottom flipped
            flipHoriz : True or *False*
                If True then the movie will be right-left flipped
            volume :
                The nominal level is 1.0, and 0.0 is silence,
                see pyglet.media.Player
            loop : bool, optional
                Whether to start the movie over from the beginning
                if draw is called and the movie is done.

        """
        # what local vars are defined (these are the init params) for use by
        # __repr__
        self._initParams = dir()
        self._initParams.remove('self')

        super(MovieStim, self).__init__(
            win, units=units, name=name, autoLog=False)
        self._verticesBase *= numpy.array([[-1, 1]])  # unflip

        if not havePygletMedia:
            msg = ("pyglet.media is needed for MovieStim and could not be"
                   " imported.\nThis can occur for various reasons;"
                   "    - psychopy.visual was imported too late (after a lib"
                   " that uses scipy)"
                   "    - no audio output is enabled (no audio card or no "
                   "speakers attached)"
                   "    - avbin is not installed")
            raise ImportError(msg)
        self._movie = None  # the actual pyglet media object
        self._player = pyglet.media.ManagedSoundPlayer()
        self._player.volume = volume
        try:
            self._player_default_on_eos = self._player.on_eos
        except Exception:
            # pyglet 1.1.4?
            self._player_default_on_eos = self._player._on_eos

        self.filename = pathToString(filename)
        self.duration = None
        self.loop = loop
        if loop and pyglet.version >= '1.2':
            logging.error("looping of movies is not currently supported "
                          "for pyglet >= 1.2 (only for version 1.1.4)")
        self.loadMovie(self.filename)
        self.format = self._movie.video_format
        self.pos = numpy.asarray(pos, float)
        self.depth = depth
        self.flipVert = flipVert
        self.flipHoriz = flipHoriz
        self.opacity = float(opacity)
        self.status = NOT_STARTED

        # size
        if size is None:
            self.size = numpy.array([self.format.width, self.format.height],
                                    float)
        else:
            self.size = val2array(size)

        self.ori = ori
        self._updateVertices()

        if win.winType != 'pyglet':
            logging.error('Movie stimuli can only be used with a '
                          'pyglet window')
            core.quit()

        # set autoLog now that params have been initialised
        wantLog = autoLog is None and self.win.autoLog
        self.__dict__['autoLog'] = autoLog or wantLog
        if self.autoLog:
            logging.exp("Created %s = %s" % (self.name, str(self)))

    def setMovie(self, filename, log=None):
        """See `~MovieStim.loadMovie` (the functions are identical).
        This form is provided for syntactic consistency with other visual
        stimuli.
        """
        self.loadMovie(filename, log=log)

    def loadMovie(self, filename, log=None):
        """Load a movie from file

        :Parameters:

            filename: string
                The name of the file, including path if necessary

        Brings up a warning if avbin is not found on the computer.
        After the file is loaded MovieStim.duration is updated with the movie
        duration (in seconds).
        """
        filename = pathToString(filename)
        try:
            self._movie = pyglet.media.load(filename, streaming=True)
        except Exception as e:
            # pyglet.media.riff is N/A if avbin is available, and then
            # actual exception would get masked with a new one for unknown
            # (sub)module riff, thus catching any exception and tuning msg
            # up if it has to do anything with avbin
            estr = str(e)
            msg = ''
            if "avbin" in estr.lower():
                msg = ("\nIt seems that avbin was not installed correctly."
                       "\nPlease fetch/install it from "
                       "http://code.google.com/p/avbin/.")
            emsg = "Caught exception '%s' while loading file '%s'.%s"
            raise IOError(emsg % (estr, filename, msg))
        self._player.queue(self._movie)
        self.duration = self._movie.duration
        while self._player.source != self._movie:
            next(self._player)
        self.status = NOT_STARTED
        self._player.pause()  # start 'playing' on the next draw command
        self.filename = filename
        logAttrib(self, log, 'movie', filename)

    def pause(self, log=None):
        """Pause the current point in the movie (sound will stop, current
        frame will not advance). If play() is called again both will restart.
        """
        self._player.pause()
        self._player._on_eos = self._player_default_on_eos
        self.status = PAUSED
        if log or log is None and self.autoLog:
            self.win.logOnFlip("Set %s paused" % self.name,
                               level=logging.EXP, obj=self)

    def stop(self, log=None):
        """Stop the current point in the movie.

        The sound will stop, current frame will not advance. Once stopped
        the movie cannot be restarted - it must be loaded again.
        Use pause() if you may need to restart the movie.
        """
        self._player.stop()
        self._player._on_eos = self._player_default_on_eos
        self.status = STOPPED
        if log or log is None and self.autoLog:
            self.win.logOnFlip("Set %s stopped" % self.name,
                               level=logging.EXP, obj=self)

    def play(self, log=None):
        """Continue a paused movie from current position.
        """
        self._player.play()
        self._player._on_eos = self._onEos
        self.status = PLAYING
        if log or log is None and self.autoLog:
            self.win.logOnFlip("Set %s playing" % self.name,
                               level=logging.EXP, obj=self)

    def seek(self, timestamp, log=None):
        """Seek to a particular timestamp in the movie.

        NB this does not seem very robust as at version 1.62, may crash!
        """
        self._player.seek(float(timestamp))
        logAttrib(self, log, 'seek', timestamp)

    def setFlipHoriz(self, newVal=True, log=None):
        """If set to True then the movie will be flipped horizontally
        (left-to-right). Note that this is relative to the original,
        not relative to the current state.
        """
        self.flipHoriz = newVal
        logAttrib(self, log, 'flipHoriz')
        self._needVertexUpdate = True

    def setFlipVert(self, newVal=True, log=None):
        """If set to True then the movie will be flipped vertically
        (top-to-bottom). Note that this is relative to the original,
        not relative to the current state.
        """
        self.flipVert = newVal
        logAttrib(self, log, 'flipVert')
        self._needVertexUpdate = True

    def draw(self, win=None):
        """Draw the current frame to a particular visual.Window.

        Draw to the default win for this object if not specified.
        The current position in the movie will be determined automatically.

        This method should be called on every frame that the movie is
        meant to appear.
        """

        if self.status == PLAYING and not self._player.playing:
            self.status = FINISHED
        _done = bool(self.status == FINISHED)
        if self.status == NOT_STARTED or (_done and self.loop):
            self.play()
        elif _done and not self.loop:
            return

        if win is None:
            win = self.win
        self._selectWindow(win)

        # make sure that textures are on and GL_TEXTURE0 is active
        GL.glActiveTexture(GL.GL_TEXTURE0)
        GL.glEnable(GL.GL_TEXTURE_2D)
        if pyglet.version >= '1.2':
            # for pyglet 1.1.4 this was done via media.dispatch_events
            self._player.update_texture()
        frameTexture = self._player.get_texture()
        if frameTexture is None:
            return

        # sets opacity (1,1,1 = RGB placeholder)
        GL.glColor4f(1, 1, 1, self.opacity)
        GL.glPushMatrix()
        self.win.setScale('pix')
        # move to centre of stimulus and rotate
        vertsPix = self.verticesPix
        t = frameTexture.tex_coords
        array = (GL.GLfloat * 32)(
            t[0], t[1],
            vertsPix[0, 0], vertsPix[0, 1], 0.,  # vertex
            t[3], t[4],
            vertsPix[1, 0], vertsPix[1, 1], 0.,
            t[6], t[7],
            vertsPix[2, 0], vertsPix[2, 1], 0.,
            t[9], t[10],
            vertsPix[3, 0], vertsPix[3, 1], 0.,
        )

        GL.glPushAttrib(GL.GL_ENABLE_BIT)
        GL.glEnable(frameTexture.target)
        GL.glBindTexture(frameTexture.target, frameTexture.id)
        GL.glPushClientAttrib(GL.GL_CLIENT_VERTEX_ARRAY_BIT)
        # 2D texture array, 3D vertex array
        GL.glInterleavedArrays(GL.GL_T2F_V3F, 0, array)
        GL.glDrawArrays(GL.GL_QUADS, 0, 4)
        GL.glPopClientAttrib()
        GL.glPopAttrib()
        GL.glPopMatrix()

    def setContrast(self):
        """Not yet implemented for MovieStim.
        """
        pass

    def _onEos(self):
        if self.loop:
            self.loadMovie(self.filename)
            self.play()
            self.status = PLAYING
        else:
            self.status = FINISHED
            self._player._on_eos = self._player_default_on_eos
        if self.autoLog:
            self.win.logOnFlip("Set %s finished" % self.name,
                               level=logging.EXP, obj=self)

    def setAutoDraw(self, val, log=None):
        """Add or remove a stimulus from the list of stimuli that will be
        automatically drawn on each flip

        :parameters:
            - val: True/False
                True to add the stimulus to the draw list, False to remove it
        """
        if val:
            self.play(log=False)  # set to play in case stopped
        else:
            self.pause(log=False)
        # add to drawing list and update status
        setAttribute(self, 'autoDraw', val, log)

    def __del__(self):
        try:
            next(self._player)
        except Exception:
            pass



# -*- coding: utf-8 -*-
"""Classes for color space pages for the color picker."""

# Part of the PsychoPy library
# Copyright (C) 2002-2018 Jonathan Peirce (C) 2019 Open Science Tools Ltd.
# Distributed under the terms of the GNU General Public License (GPL).

import wx
from psychopy.colors import Color


SLIDER_RES = 255  # resolution of the slider for color channels, leave alone!
# remember the last color and output space


class ColorPickerPageRGB(wx.Panel):
    """Class for the RGB page of the color picker.

    """
    def __init__(self,
                 parent,
                 id=wx.ID_ANY,
                 pos=wx.DefaultPosition,
                 size=wx.DefaultSize,
                 style=wx.TAB_TRAVERSAL,
                 name=wx.EmptyString):
        wx.Panel.__init__(
            self, parent, id=id, pos=pos, size=size, style=style, name=name)

        # Make sure the top-level has the correct type. We use
        # `GetTopLevelParent` to avoid creating a reference to the dialog which
        # keeps it alive and causes an access violation when PsychoPy shuts
        # down, or worse, a memory leak.
        # assert isinstance(self.GetTopLevelParent(), PsychoColorPicker)

        # Functions to convert slider units to an RGB format to display in the
        # double-spin controls beside them.
        self._posToValFunc = {0: lambda v: 2 * (v / SLIDER_RES) - 1,  # [-1:1]
                              1: lambda v: v / SLIDER_RES,  # [0:1]
                              2: lambda v: v}  # [0:255]

        # inverse of the above functions, converts values to positions
        self._valToPosFunc = {0: lambda p: int(SLIDER_RES * (p + 1) / 2.),  # [-1:1]
                              1: lambda p: int(p * SLIDER_RES),  # [0:1]
                              2: lambda p: int(p)}  # [0:255]

        self._initUI()  # setup the UI controls
        self.updateChannels()
        self.updateHex()

    def _initUI(self):
        """Initialize window controls. Called once when the page is created.

        """
        szrRGBPage = wx.BoxSizer(wx.VERTICAL)

        fraRGBChannels = wx.StaticBoxSizer(
            wx.StaticBox(self, wx.ID_ANY, u"RGBA Channels"), wx.VERTICAL)

        szrRGBArea = wx.FlexGridSizer(4, 3, 5, 5)
        szrRGBArea.AddGrowableCol(1)
        szrRGBArea.SetFlexibleDirection(wx.BOTH)
        szrRGBArea.SetNonFlexibleGrowMode(wx.FLEX_GROWMODE_SPECIFIED)

        # labels for color channels
        self.lblRed = wx.StaticText(
            fraRGBChannels.GetStaticBox(),
            wx.ID_ANY,
            u"R:",
            wx.DefaultPosition,
            wx.DefaultSize, 0)
        self.lblGreen = wx.StaticText(
            fraRGBChannels.GetStaticBox(),
            wx.ID_ANY,
            u"G:",
            wx.DefaultPosition,
            wx.DefaultSize, 0)
        self.lblBlue = wx.StaticText(
            fraRGBChannels.GetStaticBox(),
            wx.ID_ANY,
            u"B:",
            wx.DefaultPosition,
            wx.DefaultSize, 0)
        self.lblAlpha = wx.StaticText(
            fraRGBChannels.GetStaticBox(),
            wx.ID_ANY,
            u"A:",
            wx.DefaultPosition,
            wx.DefaultSize, 0)

        self.lblRed.Wrap(-1)
        self.lblGreen.Wrap(-1)
        self.lblBlue.Wrap(-1)
        self.lblAlpha.Wrap(-1)

        # sliders for setting each channel
        self.sldRed = wx.Slider(
            fraRGBChannels.GetStaticBox(),
            wx.ID_ANY,
            int(SLIDER_RES / 2), 0, SLIDER_RES,  # value, min, max
            wx.DefaultPosition,
            wx.DefaultSize,
            wx.SL_HORIZONTAL)
        self.sldGreen = wx.Slider(
            fraRGBChannels.GetStaticBox(),
            wx.ID_ANY,
            int(SLIDER_RES / 2), 0, SLIDER_RES,
            wx.DefaultPosition,
            wx.DefaultSize,
            wx.SL_HORIZONTAL)
        self.sldBlue = wx.Slider(
            fraRGBChannels.GetStaticBox(),
            wx.ID_ANY,
            int(SLIDER_RES / 2), 0, SLIDER_RES,
            wx.DefaultPosition,
            wx.DefaultSize,
            wx.SL_HORIZONTAL)
        self.sldAlpha = wx.Slider(
            fraRGBChannels.GetStaticBox(),
            wx.ID_ANY,
            SLIDER_RES, 0, SLIDER_RES,
            wx.DefaultPosition,
            wx.DefaultSize,
            wx.SL_HORIZONTAL)

        # spin (double) controls
        self.spnRed = wx.SpinCtrlDouble(
            fraRGBChannels.GetStaticBox(),
            wx.ID_ANY,
            u"0",
            wx.DefaultPosition,
            wx.DefaultSize, wx.SP_ARROW_KEYS,
            -1, 1, 0, 0.05)  # min, max, value, inc
        self.spnGreen = wx.SpinCtrlDouble(
            fraRGBChannels.GetStaticBox(),
            wx.ID_ANY,
            u"0",
            wx.DefaultPosition,
            wx.DefaultSize, wx.SP_ARROW_KEYS,
            -1, 1, 0, 0.05)
        self.spnBlue = wx.SpinCtrlDouble(
            fraRGBChannels.GetStaticBox(),
            wx.ID_ANY,
            u"0",
            wx.DefaultPosition,
            wx.DefaultSize, wx.SP_ARROW_KEYS,
            -1, 1, 0, 0.05)
        self.spnAlpha = wx.SpinCtrlDouble(
            fraRGBChannels.GetStaticBox(),
            wx.ID_ANY,
            u"1",
            wx.DefaultPosition,
            wx.DefaultSize, wx.SP_ARROW_KEYS,
            0, 1, 0, 0.05)  # non-standard specification for alpha here!!!

        self.spnRed.SetDigits(4)
        self.spnGreen.SetDigits(4)
        self.spnBlue.SetDigits(4)
        self.spnAlpha.SetDigits(4)

        # add widgets to the color channel area
        szrRGBArea.Add(
            self.lblRed, 0, wx.ALIGN_CENTER_VERTICAL | wx.ALIGN_RIGHT, 5)
        szrRGBArea.Add(self.sldRed, 1, wx.EXPAND | wx.ALIGN_CENTER_VERTICAL, 5)
        szrRGBArea.Add(self.spnRed, 0, wx.EXPAND | wx.ALIGN_CENTER_VERTICAL, 5)

        szrRGBArea.Add(
            self.lblGreen, 0, wx.ALIGN_CENTER_VERTICAL | wx.ALIGN_RIGHT, 5)
        szrRGBArea.Add(self.sldGreen, 1, wx.EXPAND | wx.ALIGN_CENTER_VERTICAL, 5)
        szrRGBArea.Add(self.spnGreen, 0, wx.EXPAND | wx.ALIGN_CENTER_VERTICAL, 5)

        szrRGBArea.Add(
            self.lblBlue, 0, wx.ALIGN_CENTER_VERTICAL | wx.ALIGN_RIGHT, 5)
        szrRGBArea.Add(self.sldBlue, 1, wx.EXPAND | wx.ALIGN_CENTER_VERTICAL, 5)
        szrRGBArea.Add(self.spnBlue, 0, wx.EXPAND | wx.ALIGN_CENTER_VERTICAL, 5)

        szrRGBArea.Add(
            self.lblAlpha, 0, wx.ALIGN_CENTER_VERTICAL | wx.ALIGN_RIGHT, 5)
        szrRGBArea.Add(
            self.sldAlpha, 1, wx.EXPAND | wx.ALIGN_CENTER_VERTICAL, 5)
        szrRGBArea.Add(self.spnAlpha, 0, wx.EXPAND | wx.ALIGN_CENTER_VERTICAL, 5)

        fraRGBChannels.Add(szrRGBArea, 1, wx.ALL | wx.EXPAND, 5)
        szrRGBPage.Add(
            fraRGBChannels, 0, wx.TOP | wx.LEFT | wx.RIGHT | wx.EXPAND, 5)

        szrRGBOptions = wx.BoxSizer(wx.HORIZONTAL)
        fraHexRGB = wx.StaticBoxSizer(
            wx.StaticBox(
                self,
                wx.ID_ANY,
                u"Hex/HTML (no Alpha)"),
            wx.VERTICAL)

        self.txtHex = wx.TextCtrl(
            fraHexRGB.GetStaticBox(),
            wx.ID_ANY,
            wx.EmptyString,
            wx.DefaultPosition,
            wx.DefaultSize,
            wx.TE_PROCESS_ENTER)
        fraHexRGB.Add(self.txtHex, 0, wx.ALL | wx.EXPAND, 5)

        szrRGBOptions.Add(fraHexRGB, 2, 0, 5)

        # the index of each item maps to the conversion functions
        rbxRGBModeChoices = [
            u"PsychoPy RGB [-1:1]",
            u"Normalized RGB [0:1]",
            u"8-Bit RGB [0:255]"]
        self.rbxRGBFormat = wx.RadioBox(
            self,
            wx.ID_ANY,
            u"RGB Format",
            wx.DefaultPosition,
            wx.DefaultSize,
            rbxRGBModeChoices,
            1, wx.RA_SPECIFY_COLS)
        self.rbxRGBFormat.SetSelection(0)
        # self.rbxRGBFormat.SetToolTip(
        #     u"Format to use for specifying RGBA channels.")

        szrRGBOptions.Add(self.rbxRGBFormat, 0, wx.LEFT, 5)

        szrRGBPage.Add(szrRGBOptions, 1, wx.EXPAND | wx.ALL, 5)

        self.SetSizer(szrRGBPage)
        self.Layout()

        # Connect Events
        self.sldRed.Bind(wx.EVT_SCROLL, self.onRedScroll)
        self.spnRed.Bind(wx.EVT_SPINCTRLDOUBLE, self.onRedUpdate)
        self.spnRed.Bind(wx.EVT_TEXT_ENTER, self.onRedUpdate)
        self.sldGreen.Bind(wx.EVT_SCROLL, self.onGreenScroll)
        self.spnGreen.Bind(wx.EVT_SPINCTRLDOUBLE, self.onGreenUpdate)
        self.spnGreen.Bind(wx.EVT_TEXT_ENTER, self.onGreenUpdate)
        self.sldBlue.Bind(wx.EVT_SCROLL, self.onBlueScroll)
        self.spnBlue.Bind(wx.EVT_SPINCTRLDOUBLE, self.onBlueUpdate)
        self.spnBlue.Bind(wx.EVT_TEXT_ENTER, self.onBlueUpdate)
        self.sldAlpha.Bind(wx.EVT_SCROLL, self.onAlphaScroll)
        self.spnAlpha.Bind(wx.EVT_SPINCTRLDOUBLE, self.onAlphaUpdate)
        self.spnAlpha.Bind(wx.EVT_TEXT_ENTER, self.onAlphaUpdate)
        self.txtHex.Bind(wx.EVT_TEXT_ENTER, self.onHexChanged)
        self.rbxRGBFormat.Bind(wx.EVT_RADIOBOX, self.onRGBMode)

    def updateDialog(self):
        """Update the color specified by the dialog.

        """
        # update the color specified by the dialog
        dlg = self.GetTopLevelParent()
        dlgColor = self.GetTopLevelParent().color

        spinVals = [
            self.spnRed.GetValue(),
            self.spnGreen.GetValue(),
            self.spnBlue.GetValue(),
            self.spnAlpha.GetValue()]

        channelFormat = self.rbxRGBFormat.GetSelection()
        if channelFormat == 0:
            dlgColor.rgba = spinVals
        elif channelFormat == 1:
            dlgColor.rgba1 = spinVals
        elif channelFormat == 2:
            dlgColor.rgba255 = spinVals
        else:
            raise ValueError('Color channel format not supported.')

        if hasattr(dlg, 'pnlColorPreview'):
            dlg.pnlColorPreview.Refresh()

    def updateChannels(self):
        """Update the values of the channels to reflect the color specified by
        the dialog. This should never be called within any event handler for
        controls within the channels frame!

        """
        # get colors and convert to format wxPython controls can accept
        rgbaColor = self.GetTopLevelParent().color
        rgba255 = [int(i) for i in rgbaColor.rgba255]
        self.sldRed.SetValue(rgba255[0])
        self.sldGreen.SetValue(rgba255[1])
        self.sldBlue.SetValue(rgba255[2])
        self.sldAlpha.SetValue(rgbaColor.alpha * SLIDER_RES)  # arrrg! should be 255!!!

        channelMode = self.rbxRGBFormat.GetSelection()
        convFunc = self._posToValFunc[channelMode]

        # update spinner values/ranges for each channel
        for spn in (self.spnRed, self.spnGreen, self.spnBlue):
            spn.SetDigits(
                0 if self.rbxRGBFormat.GetSelection() == 2 else 4)
            spn.SetIncrement(
                1 if self.rbxRGBFormat.GetSelection() == 2 else 0.05)
            spn.SetMin(convFunc(0))
            spn.SetMax(convFunc(SLIDER_RES))

        # set the value in the new range

        if channelMode == 0:
            spnColVals = rgbaColor.rgba
        elif channelMode == 1:
            spnColVals = rgbaColor.rgba1
        elif channelMode == 2:
            spnColVals = rgbaColor.rgba255
        else:
            raise ValueError("Unknown RGB channel format specified. Did you add"
                             "items to `rbxRGBFormat`?")

        self.spnRed.SetValue(spnColVals[0])
        self.spnGreen.SetValue(spnColVals[1])
        self.spnBlue.SetValue(spnColVals[2])
        self.spnAlpha.SetValue(rgbaColor.alpha)

    def onRedScroll(self, event):
        """Called when the red channel slider is moved. Updates the spin control
        and the color specified by the dialog.

        """
        self.spnRed.SetValue(
            self._posToValFunc[self.rbxRGBFormat.GetSelection()](
                event.Position))
        self.updateHex()
        self.updateDialog()
        event.Skip()

    def onRedUpdate(self, event):
        """Called when the red channel spin control is changed. Updates the hex
        value and the color specified by the dialog.

        """
        self.sldRed.SetValue(
            self._valToPosFunc[self.rbxRGBFormat.GetSelection()](event.Value))
        self.updateHex()
        self.updateDialog()
        event.Skip()

    def onGreenScroll(self, event):
        """Called when the green channel slider is moved. Updates the spin
        control and the color specified by the dialog.

        """
        self.spnGreen.SetValue(
            self._posToValFunc[self.rbxRGBFormat.GetSelection()](
                event.Position))
        self.updateHex()
        self.updateDialog()
        event.Skip()

    def onGreenUpdate(self, event):
        """Called when the green channel spin control is changed. Updates the
        hex value and the color specified by the dialog.

        """
        self.sldGreen.SetValue(
            self._valToPosFunc[self.rbxRGBFormat.GetSelection()](event.Value))
        self.updateHex()
        self.updateDialog()
        event.Skip()

    def onBlueScroll(self, event):
        """Called when the blue channel slider is moved. Updates the spin
        control and the color specified by the dialog.

        """
        self.spnBlue.SetValue(
            self._posToValFunc[self.rbxRGBFormat.GetSelection()](
                event.Position))
        self.updateHex()
        self.updateDialog()
        event.Skip()

    def onBlueUpdate(self, event):
        """Called when the blue channel spin control is changed. Updates the hex
        value and the color specified by the dialog.

        """
        self.sldBlue.SetValue(
            self._valToPosFunc[self.rbxRGBFormat.GetSelection()](event.Value))
        self.updateHex()
        self.updateDialog()
        event.Skip()

    def onAlphaScroll(self, event):
        """Called when the alpha (transparency) channel slider is moved. Updates
        the spin control and the color specified by the dialog.

        """
        self.spnAlpha.SetValue(event.Position / SLIDER_RES)
        self.updateHex()
        self.updateDialog()
        event.Skip()

    def onAlphaUpdate(self, event):
        """Called when the alpha channel spin control is changed. Updates the
        hex value and the color specified by the dialog.

        """
        self.sldAlpha.SetValue(event.Value * SLIDER_RES)
        self.updateHex()
        self.updateDialog()
        event.Skip()

    def onHexChanged(self, event):
        """Called when the user manually enters a hex value into the field.
        If the color value is valid, the new value will appear and the channels
        will update.

        """
        dlgColor = self.GetTopLevelParent().color
        try:
            dlgColor.rgba = Color(self.txtHex.GetValue(), space='hex').rgba
        except ValueError:
            pass

        self.updateHex()
        self.updateChannels()
        self.updateDialog()
        event.Skip()

    def updateHex(self):
        """Update the hex/HTML value using the color specified by the dialog.

        """
        self.txtHex.SetValue(self.GetTopLevelParent().color.hex)

    def onRGBMode(self, event):
        """Called when the RGB mode is changed. This re-ranges all the spin
        controls and updates the interface."""
        self.updateChannels()
        self.updateHex()
        event.Skip()


class ColorPickerPageHSV(wx.Panel):
    """Class for the HSV page of the color picker.

    """
    def __init__(self,
                 parent,
                 id=wx.ID_ANY,
                 pos=wx.DefaultPosition,
                 size=wx.DefaultSize,
                 style=wx.TAB_TRAVERSAL,
                 name=wx.EmptyString):
        wx.Panel.__init__(
            self, parent, id=id, pos=pos, size=size, style=style, name=name)

        self._initUI()  # setup the UI controls
        self.updateChannels()

    def _initUI(self):
        """Initialize window controls. Called once when the page is created.

        """
        szrRGBPage = wx.BoxSizer(wx.VERTICAL)

        fraHSVChannels = wx.StaticBoxSizer(
            wx.StaticBox(self, wx.ID_ANY, u"HSV/HSB Values"), wx.VERTICAL)

        szrHSVArea = wx.FlexGridSizer(4, 3, 5, 5)
        szrHSVArea.AddGrowableCol(1)
        szrHSVArea.SetFlexibleDirection(wx.BOTH)
        szrHSVArea.SetNonFlexibleGrowMode(wx.FLEX_GROWMODE_SPECIFIED)

        # labels for color channels
        self.lblHue = wx.StaticText(
            fraHSVChannels.GetStaticBox(),
            wx.ID_ANY,
            u"H:",
            wx.DefaultPosition,
            wx.DefaultSize, 0)
        self.lblSat = wx.StaticText(
            fraHSVChannels.GetStaticBox(),
            wx.ID_ANY,
            u"S:",
            wx.DefaultPosition,
            wx.DefaultSize, 0)
        self.lblVal = wx.StaticText(
            fraHSVChannels.GetStaticBox(),
            wx.ID_ANY,
            u"V:",
            wx.DefaultPosition,
            wx.DefaultSize, 0)
        self.lblAlpha = wx.StaticText(
            fraHSVChannels.GetStaticBox(),
            wx.ID_ANY,
            u"A:",
            wx.DefaultPosition,
            wx.DefaultSize, 0)

        self.lblHue.Wrap(-1)
        self.lblSat.Wrap(-1)
        self.lblVal.Wrap(-1)
        self.lblAlpha.Wrap(-1)

        # sliders for setting each channel
        self.sldHue = wx.Slider(
            fraHSVChannels.GetStaticBox(),
            wx.ID_ANY,
            0, 0, 360,  # value, min, max
            wx.DefaultPosition,
            wx.DefaultSize,
            wx.SL_HORIZONTAL)
        self.sldSat = wx.Slider(
            fraHSVChannels.GetStaticBox(),
            wx.ID_ANY,
            0, 0, SLIDER_RES,
            wx.DefaultPosition,
            wx.DefaultSize,
            wx.SL_HORIZONTAL)
        self.sldVal = wx.Slider(
            fraHSVChannels.GetStaticBox(),
            wx.ID_ANY,
            0, 0, SLIDER_RES,
            wx.DefaultPosition,
            wx.DefaultSize,
            wx.SL_HORIZONTAL)
        self.sldAlpha = wx.Slider(
            fraHSVChannels.GetStaticBox(),
            wx.ID_ANY,
            0, 0, SLIDER_RES,
            wx.DefaultPosition,
            wx.DefaultSize,
            wx.SL_HORIZONTAL)

        # spin (double) controls
        self.spnHue = wx.SpinCtrlDouble(
            fraHSVChannels.GetStaticBox(),
            wx.ID_ANY,
            u"0",
            wx.DefaultPosition,
            wx.DefaultSize, wx.SP_ARROW_KEYS,
            0, 360, 0, 1.0)  # min, max, value, inc
        self.spnSat = wx.SpinCtrlDouble(
            fraHSVChannels.GetStaticBox(),
            wx.ID_ANY,
            u"0",
            wx.DefaultPosition,
            wx.DefaultSize, wx.SP_ARROW_KEYS,
            0, 1, 0, 0.05)
        self.spnVal = wx.SpinCtrlDouble(
            fraHSVChannels.GetStaticBox(),
            wx.ID_ANY,
            u"0",
            wx.DefaultPosition,
            wx.DefaultSize, wx.SP_ARROW_KEYS,
            0, 1, 0, 0.05)
        self.spnAlpha = wx.SpinCtrlDouble(
            fraHSVChannels.GetStaticBox(),
            wx.ID_ANY,
            u"1",
            wx.DefaultPosition,
            wx.DefaultSize, wx.SP_ARROW_KEYS,
            0, 1, 0, 0.05)

        self.spnHue.SetDigits(4)
        self.spnSat.SetDigits(4)
        self.spnVal.SetDigits(4)
        self.spnAlpha.SetDigits(4)

        # add widgets to the color channel area
        szrHSVArea.Add(
            self.lblHue, 0, wx.ALIGN_CENTER_VERTICAL | wx.ALIGN_RIGHT, 5)
        szrHSVArea.Add(self.sldHue, 1, wx.EXPAND, 5)
        szrHSVArea.Add(self.spnHue, 0, wx.EXPAND | wx.ALIGN_CENTER_VERTICAL, 5)

        szrHSVArea.Add(
            self.lblSat, 0, wx.ALIGN_CENTER_VERTICAL | wx.ALIGN_RIGHT, 5)
        szrHSVArea.Add(self.sldSat, 1, wx.EXPAND, 5)
        szrHSVArea.Add(self.spnSat, 0, wx.EXPAND | wx.ALIGN_CENTER_VERTICAL, 5)

        szrHSVArea.Add(
            self.lblVal, 0, wx.ALIGN_CENTER_VERTICAL | wx.ALIGN_RIGHT, 5)
        szrHSVArea.Add(self.sldVal, 10, wx.EXPAND, 5)
        szrHSVArea.Add(self.spnVal, 0, wx.EXPAND | wx.ALIGN_CENTER_VERTICAL, 5)

        szrHSVArea.Add(
            self.lblAlpha, 0, wx.ALIGN_CENTER_VERTICAL | wx.ALIGN_RIGHT, 5)
        szrHSVArea.Add(
            self.sldAlpha, 1, wx.EXPAND | wx.ALIGN_CENTER_VERTICAL, 5)
        szrHSVArea.Add(self.spnAlpha, 0, wx.EXPAND | wx.ALIGN_CENTER_VERTICAL, 5)

        fraHSVChannels.Add(szrHSVArea, 1, wx.ALL | wx.EXPAND, 5)
        szrRGBPage.Add(
            fraHSVChannels, 0, wx.TOP | wx.LEFT | wx.RIGHT | wx.EXPAND, 5)

        self.SetSizer(szrRGBPage)
        self.Layout()

        # Connect Events
        self.sldHue.Bind(wx.EVT_SCROLL, self.onHueScroll)
        self.spnHue.Bind(wx.EVT_SPINCTRLDOUBLE, self.onHueUpdate)
        self.spnHue.Bind(wx.EVT_TEXT_ENTER, self.onHueUpdate)
        self.sldSat.Bind(wx.EVT_SCROLL, self.onSatScroll)
        self.spnSat.Bind(wx.EVT_SPINCTRLDOUBLE, self.onSatUpdate)
        self.spnSat.Bind(wx.EVT_TEXT_ENTER, self.onSatUpdate)
        self.sldVal.Bind(wx.EVT_SCROLL, self.onValScroll)
        self.spnVal.Bind(wx.EVT_SPINCTRLDOUBLE, self.onValUpdate)
        self.spnVal.Bind(wx.EVT_TEXT_ENTER, self.onValUpdate)
        self.sldAlpha.Bind(wx.EVT_SCROLL, self.onAlphaScroll)
        self.spnAlpha.Bind(wx.EVT_SPINCTRLDOUBLE, self.onAlphaUpdate)
        self.spnAlpha.Bind(wx.EVT_TEXT_ENTER, self.onAlphaUpdate)

    def updateDialog(self):
        """Update the color specified by the dialog.

        """
        # update the color specified by the dialog
        dlg = self.GetTopLevelParent()
        dlgColor = self.GetTopLevelParent().color

        spinVals = [
            self.spnHue.GetValue(),
            self.spnSat.GetValue(),
            self.spnVal.GetValue(),
            self.spnAlpha.GetValue()]

        dlgColor.hsva = spinVals

        if hasattr(dlg, 'pnlColorPreview'):
            dlg.pnlColorPreview.Refresh()

    def updateChannels(self):
        """Update the values of the channels to reflect the color specified by
        the dialog. This should never be called within any event handler for
        controls within the channels frame!

        """
        # get colors and convert to format wxPython controls can accept
        rgbaColor = self.GetTopLevelParent().color
        hsva = [i for i in rgbaColor.hsva]

        self.sldHue.SetValue(hsva[0])
        self.sldSat.SetValue(hsva[1] * SLIDER_RES)
        self.sldVal.SetValue(hsva[2] * SLIDER_RES)
        self.sldAlpha.SetValue(hsva[3] * SLIDER_RES)  # arrrg! should be 255!!!

        # set the value in the new range
        self.spnHue.SetValue(hsva[0])
        self.spnSat.SetValue(hsva[1])
        self.spnVal.SetValue(hsva[2])
        self.spnAlpha.SetValue(hsva[3])

    def onHueScroll(self, event):
        """Called when the red channel slider is moved. Updates the spin control
        and the color specified by the dialog.

        """
        self.spnHue.SetValue(event.Position)
        self.updateDialog()
        event.Skip()

    def onHueUpdate(self, event):
        """Called when the red spin control. Updates the hex value and the color
        specified by the dialog.

        """
        self.sldHue.SetValue(int(event.GetValue()))
        self.updateDialog()
        event.Skip()

    def onSatScroll(self, event):
        """Called when the green channel slider is moved. Updates the spin
        control and the color specified by the dialog.

        """
        self.spnSat.SetValue(event.Position / SLIDER_RES)
        self.updateDialog()
        event.Skip()

    def onSatUpdate(self, event):
        """Called when the green spin control. Updates the hex value and the
        color specified by the dialog.

        """
        self.sldSat.SetValue(event.GetValue() * SLIDER_RES)
        self.updateDialog()
        event.Skip()

    def onValScroll(self, event):
        """Called when the blue channel slider is moved. Updates the spin
        control and the color specified by the dialog.

        """
        self.spnVal.SetValue(event.Position / SLIDER_RES)
        self.updateDialog()
        event.Skip()

    def onValUpdate(self, event):
        """Called when the blue spin control. Updates the hex value and the
        color specified by the dialog.

        """
        self.sldVal.SetValue(event.GetValue() * SLIDER_RES)
        self.updateDialog()
        event.Skip()

    def onAlphaScroll(self, event):
        """Called when the alpha (transparency) channel slider is moved. Updates
        the spin control and the color specified by the dialog.

        """
        self.spnAlpha.SetValue(event.Position / SLIDER_RES)
        self.updateDialog()
        event.Skip()

    def onAlphaUpdate(self, event):
        """Called when the alpha spin control. Updates the hex value and the
        color specified by the dialog.

        """
        self.sldAlpha.SetValue(event.GetValue() * SLIDER_RES)
        self.updateDialog()
        event.Skip()


if __name__ == "__main__":
    pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-


"""
In PsychoPy 1.81.00 the support for Bits++ has been substantially rewritten and will
require a few changes to your code. It will allow a greater range of devices to be used
in future (including BitsSharp).

From now on, rather than the device/mode being a setting of the window you now
create the bitsBox object yourself, providing it with a reference to the window
and then interact with the box directly. So instead of
   win = visual.Window([1024, 768], bitsmode='fast')
   win.bits.setContrast(0.5)

you would now use:
   win = visual.Window([1024, 768])
   bitsBox = crs.BitsSharp(win, mode='bits++')
   bitsBox.setContrast(0.5)

Check your experiment still works as expected!
"""

from __future__ import absolute_import, division, print_function

from psychopy import visual, core, event, logging
from psychopy.hardware import crs
import numpy

# This is the new way to use the Bits++ box (or a Bits# device) with PsychoPy.
# The BitsSharp device under PsychoPy has some issues right now:
#    - the shaders aren't working for mono++and color++modes
#    - for BitsSharp, the device will try to deduce the identity LUT of the screen
#    but to do that it needs to be in fullscreen mode
# logging.console.setLevel(logging.INFO)
win = visual.Window([1024, 768], useFBO=True, fullscr=True, screen = 0)
bitsBox = crs.BitsPlusPlus(win, mode='bits++',
        rampType='configFile')  # change this to an integer to use a specific identityLUT

# BitsSharp can check identity LUT automatically:
# bitsBox = crs.BitsSharp(win, mode='bits++', checkConfigLevel=1)
# if not bitsBox.OK:
#    print('failed to connect to Bits box')
# else:
#    print('found %s on %s' %(bitsBox.type, bitsBox.com.port))

grating = visual.GratingStim(win, mask = 'gauss', ori=45, sf=2)

# Using bits++ with one stimulus
globalClock = core.Clock()
while True:
    # get new contrast
    t = globalClock.getTime()
    newContr = numpy.sin(t * numpy.pi * 2)  # sinusoidally modulate contrast

    # set whole screen to this contrast
    bitsBox.setContrast(newContr)

    # draw gratings and update screen
    grating.draw()
    win.flip()

    # check for a keypress
    if event.getKeys():
        break
    event.clearEvents('mouse')  # only really needed for pygame windows

# reset the bits++(and update the window so that this is done properly)
bitsBox.setContrast(1)
win.flip()

win.close()
core.quit()

# The contents of this file are in the public domain.



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Part of the PsychoPy library
# Copyright (C) 2002-2018 Jonathan Peirce (C) 2019-2021 Open Science Tools Ltd.
# Distributed under the terms of the GNU General Public License (GPL).

from __future__ import absolute_import, division, print_function

# from future import standard_library
# standard_library.install_aliases()
from builtins import str
from builtins import range
from past.builtins import basestring
import os
import re
import ast
import pickle
import time
import codecs
import numpy as np
import pandas as pd

from collections import OrderedDict
from pkg_resources import parse_version

from psychopy import logging, exceptions
from psychopy.constants import PY3
from psychopy.tools.filetools import pathToString

try:
    import openpyxl
    if parse_version(openpyxl.__version__) >= parse_version('2.4.0'):
        # openpyxl moved get_column_letter to utils.cell
        from openpyxl.utils.cell import get_column_letter
    else:
        from openpyxl.cell import get_column_letter
    from openpyxl.reader.excel import load_workbook
    haveOpenpyxl = True
except ImportError:
    haveOpenpyxl = False

haveXlrd = False

_nonalphanumeric_re = re.compile(r'\W')  # will match all bad var name chars


def checkValidFilePath(filepath, makeValid=True):
    """Checks whether file path location (e.g. is a valid folder)

    This should also check whether we have write-permissions to the folder
    but doesn't currently do that!

    added in: 1.90.00
    """
    folder = os.path.split(os.path.abspath(filepath))[0]
    if not os.path.isdir(folder):
        os.makedirs(folder)  # spit an error if we fail
    return True


def isValidVariableName(name):
    """Checks whether a certain string could be used as a valid variable.

    Usage::

        OK, msg = isValidVariableName(name)

    >>> isValidVariableName('name')
    (True, '')
    >>> isValidVariableName('0name')
    (False, 'Variables cannot begin with numeric character')
    >>> isValidVariableName('first second')
    (False, 'Variables cannot contain punctuation or spaces')
    >>> isValidVariableName('')
    (False, "Variables cannot be missing, None, or ''")
    >>> isValidVariableName(None)
    (False, "Variables cannot be missing, None, or ''")
    >>> isValidVariableName(23)
    (False, "Variables must be string-like")
    >>> isValidVariableName('a_b_c')
    (True, '')
    """
    if not name:
        return False, "Variables cannot be missing, None, or ''"
    if not isinstance(name, basestring):
        return False, "Variables must be string-like"
    try:
        name = str(name)  # convert from unicode if possible
    except Exception:
        if type(name) in [str, np.unicode_]:
            msg = ("name %s (type %s) contains non-ASCII characters"
                   " (e.g. accents)")
            raise AttributeError(msg % (name, type(name)))
        else:
            msg = "name %s (type %s) could not be converted to a string"
            raise AttributeError(msg % (name, type(name)))

    if name[0].isdigit():
        return False, "Variables cannot begin with numeric character"
    if _nonalphanumeric_re.search(name):
        return False, "Variables cannot contain punctuation or spaces"
    return True, ''


def _getExcelCellName(col, row):
    """Returns the excel cell name for a row and column (zero-indexed)

    >>> _getExcelCellName(0,0)
    'A1'
    >>> _getExcelCellName(2,1)
    'C2'
    """
    # BEWARE - openpyxl uses indexing at 1, to fit with Excel
    return "%s%i" % (get_column_letter(col + 1), row + 1)


def importTrialTypes(fileName, returnFieldNames=False):
    """importTrialTypes is DEPRECATED (as of v1.70.00)
    Please use `importConditions` for identical functionality.
    """
    logging.warning("importTrialTypes is DEPRECATED (as of v1.70.00). "
                    "Please use `importConditions` for identical "
                    "functionality.")
    return importConditions(fileName, returnFieldNames)


def sliceFromString(sliceString):
    """Convert a text string into a valid slice object
    which can be used as indices for a list or array.

    >>> sliceFromString("0:10")
    slice(0,10,None)
    >>> sliceFromString("0::3")
    slice(0,None,3)
    >>> sliceFromString("-8:")
    slice(-8,None,None)
    """
    sliceArgs = []
    for val in sliceString.split(':'):
        if len(val) == 0:
            sliceArgs.append(None)
        else:
            sliceArgs.append(int(round(float(val))))
            # nb int(round(float(x))) is needed for x='4.3'
    return slice(*sliceArgs)


def indicesFromString(indsString):
    """Convert a text string into a valid list of indices
    """
    # "6"
    try:
        inds = int(round(float(indsString)))
        return [inds]
    except Exception:
        pass
    # "-6::2"
    try:
        inds = sliceFromString(indsString)
        return inds
    except Exception:
        pass
    # "1,4,8"
    try:
        inds = list(eval(indsString))
        return inds
    except Exception:
        pass


def listFromString(val, excludeEmpties=False):
    """Take a string that looks like a list (with commas and/or [] and make
    an actual python list"""
    # was previously called strToList and str2list might have been an option!
    # I'll leave those here for anyone doing a find-in-path for those
    if type(val) == tuple:
        return list(val)
    elif type(val) == list:
        return list(val)  # nothing to do
    elif type(val) != str:
        raise ValueError("listFromString requires a string as its input not {}"
                         .format(repr(val)))
    # try to evaluate with ast (works for "'yes,'no'" or "['yes', 'no']")
    try:
        iterable = ast.literal_eval(val)
        if type(iterable) == tuple:
            iterable = list(iterable)
        return iterable
    except (ValueError, SyntaxError):
        pass  # e.g. "yes, no" won't work. We'll go on and try another way

    val = val.strip()  # in case there are spaces
    if val.startswith(('[', '(')) and val.endswith((']', ')')):
        val = val[1:-1]
    asList = val.split(",")
    if excludeEmpties:
        asList = [this.strip() for this in asList if this]
    else:
        asList = [this.strip() for this in asList]
    return asList


def importConditions(fileName, returnFieldNames=False, selection=""):
    """Imports a list of conditions from an .xlsx, .csv, or .pkl file

    The output is suitable as an input to :class:`TrialHandler`
    `trialList` or to :class:`MultiStairHandler` as a `conditions` list.

    If `fileName` ends with:

        - .csv:  import as a comma-separated-value file
            (header + row x col)
        - .xlsx: import as Excel 2007 (xlsx) files.
            No support for older (.xls) is planned.
        - .pkl:  import from a pickle file as list of lists
            (header + row x col)

    The file should contain one row per type of trial needed and one column
    for each parameter that defines the trial type. The first row should give
    parameter names, which should:

        - be unique
        - begin with a letter (upper or lower case)
        - contain no spaces or other punctuation (underscores are permitted)


    `selection` is used to select a subset of condition indices to be used
    It can be a list/array of indices, a python `slice` object or a string to
    be parsed as either option.
    e.g.:

        - "1,2,4" or [1,2,4] or (1,2,4) are the same
        - "2:5"       # 2, 3, 4 (doesn't include last whole value)
        - "-10:2:"    # tenth from last to the last in steps of 2
        - slice(-10, 2, None)  # the same as above
        - random(5) * 8  # five random vals 0-7

    """

    def _attemptImport(fileName, sep=',', dec='.'):
        """Attempts to import file with specified settings and raises
        ConditionsImportError if fails due to invalid format

        :param filename: str
        :param sep: str indicating the separator for cells (',', ';' etc)
        :param dec: str indicating the decimal point ('.', '.')
        :return: trialList, fieldNames
        """
        if fileName.endswith(('.csv', '.tsv')):
            trialsArr = pd.read_csv(fileName, encoding='utf-8-sig',
                                    sep=sep, decimal=dec)
            for col in trialsArr.columns:
                for row, cell in enumerate(trialsArr[col]):
                    if isinstance(cell, str):
                        tryVal = cell.replace(",", ".")
                        try:
                            trialsArr[col][row] = float(tryVal)
                        except ValueError:
                            pass
            logging.debug(u"Read csv file with pandas: {}".format(fileName))
        elif fileName.endswith(('.xlsx', '.xlsm')):
            trialsArr = pd.read_excel(fileName, engine='openpyxl')
            logging.debug(u"Read Excel file with pandas: {}".format(fileName))
        elif fileName.endswith('.xls'):
            trialsArr = pd.read_excel(fileName, engine='xlrd')
            logging.debug(u"Read Excel file with pandas: {}".format(fileName))
        # then try to convert array to trialList and fieldnames
        unnamed = trialsArr.columns.to_series().str.contains('^Unnamed: ')
        trialsArr = trialsArr.loc[:, ~unnamed]  # clear unnamed cols
        logging.debug(u"Clearing unnamed columns from {}".format(fileName))
        trialList, fieldNames = pandasToDictList(trialsArr)

        return trialList, fieldNames

    def _assertValidVarNames(fieldNames, fileName):
        """screens a list of names as candidate variable names. if all
        names are OK, return silently; else raise  with msg
        """
        fileName = pathToString(fileName)
        if not all(fieldNames):
            msg = ('Conditions file %s: Missing parameter name(s); '
                   'empty cell(s) in the first row?')
            raise exceptions.ConditionsImportError(msg % fileName)
        for name in fieldNames:
            OK, msg = isValidVariableName(name)
            if not OK:
                # tailor message to importConditions
                msg = msg.replace('Variables', 'Parameters (column headers)')
                raise exceptions.ConditionsImportError(
                    'Conditions file %s: %s%s"%s"' %
                    (fileName, msg, os.linesep * 2, name))

    if fileName in ['None', 'none', None]:
        if returnFieldNames:
            return [], []
        return []
    if not os.path.isfile(fileName):
        msg = 'Conditions file not found: %s'
        raise ValueError(msg % os.path.abspath(fileName))

    def pandasToDictList(dataframe):
        """Convert a pandas dataframe to a list of dicts.
        This helper function is used by csv or excel imports via pandas
        """
        # convert the resulting dataframe to a numpy recarray
        trialsArr = dataframe.to_records(index=False)
        # Check for new line characters in strings, and replace escaped characters
        for record in trialsArr:
            for idx, element in enumerate(record):
                if isinstance(element, str):
                    record[idx] = element.replace('\\n', '\n')
        if trialsArr.shape == ():
            # convert 0-D to 1-D with one element:
            trialsArr = trialsArr[np.newaxis]
        fieldNames = list(trialsArr.dtype.names)
        _assertValidVarNames(fieldNames, fileName)

        # convert the record array into a list of dicts
        trialList = []
        for trialN, trialType in enumerate(trialsArr):
            thisTrial = OrderedDict()
            for fieldN, fieldName in enumerate(fieldNames):
                val = trialsArr[trialN][fieldN]

                if isinstance(val, basestring):
                    if val.startswith('[') and val.endswith(']'):
                        # val = eval('%s' %unicode(val.decode('utf8')))
                        val = eval(val)
                elif type(val) == np.string_:
                    val = str(val.decode('utf-8-sig'))
                    # if it looks like a list, convert it:
                    if val.startswith('[') and val.endswith(']'):
                        # val = eval('%s' %unicode(val.decode('utf8')))
                        val = eval(val)
                elif np.isnan(val):
                    val = None
                thisTrial[fieldName] = val
            trialList.append(thisTrial)
        return trialList, fieldNames

    if (fileName.endswith(('.csv', '.tsv'))
            or (fileName.endswith(('.xlsx', '.xls', '.xlsm')) and haveXlrd)):
        if fileName.endswith(('.csv', '.tsv', '.dlm')):  # delimited text file
            for sep, dec in [ (',', '.'), (';', ','),  # most common in US, EU
                              ('\t', '.'), ('\t', ','), (';', '.')]:
                try:
                    trialList, fieldNames = _attemptImport(fileName=fileName,
                                                           sep=sep, dec=dec)
                    break  # seems to have worked
                except exceptions.ConditionsImportError as e:
                    continue  # try a different format
        else:
            trialList, fieldNames = _attemptImport(fileName=fileName)

    elif fileName.endswith(('.xlsx','.xlsm')):  # no xlsread so use openpyxl
        if not haveOpenpyxl:
            raise ImportError('openpyxl or xlrd is required for loading excel '
                              'files, but neither was found.')

        # data_only was added in 1.8
        if parse_version(openpyxl.__version__) < parse_version('1.8'):
            wb = load_workbook(filename=fileName)
        else:
            wb = load_workbook(filename=fileName, data_only=True)
        ws = wb.worksheets[0]

        logging.debug(u"Read excel file with openpyxl: {}".format(fileName))
        try:
            # in new openpyxl (2.3.4+) get_highest_xx is deprecated
            nCols = ws.max_column
            nRows = ws.max_row
        except Exception:
            # version openpyxl 1.5.8 (in Standalone 1.80) needs this
            nCols = ws.get_highest_column()
            nRows = ws.get_highest_row()

        # get parameter names from the first row header
        fieldNames = []
        rangeCols = list(range(nCols))
        for colN in range(nCols):
            if parse_version(openpyxl.__version__) < parse_version('2.0'):
                fieldName = ws.cell(_getExcelCellName(col=colN, row=0)).value
            else:
                # From 2.0, cells are referenced with 1-indexing: A1 == cell(row=1, column=1)
                fieldName = ws.cell(row=1, column=colN + 1).value
            if fieldName:
                # If column is named, add its name to fieldNames
                fieldNames.append(fieldName)
            else:
                # Otherwise, ignore the column
                rangeCols.remove(colN)
        _assertValidVarNames(fieldNames, fileName)

        # loop trialTypes
        trialList = []
        for rowN in range(1, nRows):  # skip header first row
            thisTrial = {}
            for colN in rangeCols:
                if parse_version(openpyxl.__version__) < parse_version('2.0'):
                    val = ws.cell(_getExcelCellName(col=colN, row=0)).value
                else:
                    # From 2.0, cells are referenced with 1-indexing: A1 == cell(row=1, column=1)
                    val = ws.cell(row=rowN + 1, column=colN + 1).value
                # if it looks like a list or tuple, convert it
                if (isinstance(val, basestring) and
                        (val.startswith('[') and val.endswith(']') or
                                 val.startswith('(') and val.endswith(')'))):
                    val = eval(val)
                # Convert from eu style decimals: replace , with . and try to make it a float
                if isinstance(val, basestring):
                    tryVal = val.replace(",", ".")
                    try:
                        val = float(tryVal)
                    except ValueError:
                        pass
                fieldName = fieldNames[colN]
                thisTrial[fieldName] = val
            trialList.append(thisTrial)

    elif fileName.endswith('.pkl'):
        f = open(fileName, 'rb')
        # Converting newline characters.
        if PY3:
            # 'b' is necessary in Python3 because byte object is 
            # returned when file is opened in binary mode.
            buffer = f.read().replace(b'\r\n',b'\n').replace(b'\r',b'\n')
        else:
            buffer = f.read().replace('\r\n','\n').replace('\r','\n')
        try:
            trialsArr = pickle.loads(buffer)
        except Exception:
            raise IOError('Could not open %s as conditions' % fileName)
        f.close()
        trialList = []
        if PY3:
            # In Python3, strings returned by pickle() is unhashable.
            # So, we have to convert them to str.
            trialsArr = [[str(item) if isinstance(item, str) else item
                          for item in row] for row in trialsArr]
        fieldNames = trialsArr[0]  # header line first
        _assertValidVarNames(fieldNames, fileName)
        for row in trialsArr[1:]:
            thisTrial = {}
            for fieldN, fieldName in enumerate(fieldNames):
                # type is correct, being .pkl
                thisTrial[fieldName] = row[fieldN]
            trialList.append(thisTrial)
    else:
        raise IOError('Your conditions file should be an '
                      'xlsx, csv, dlm, tsv or pkl file')

    # if we have a selection then try to parse it
    if isinstance(selection, basestring) and len(selection) > 0:
        selection = indicesFromString(selection)
        if not isinstance(selection, slice):
            for n in selection:
                try:
                    assert n == int(n)
                except AssertionError:
                    raise TypeError("importConditions() was given some "
                                    "`indices` but could not parse them")

    # the selection might now be a slice or a series of indices
    if isinstance(selection, slice):
        trialList = trialList[selection]
    elif len(selection) > 0:
        allConds = trialList
        trialList = []
        print(selection)
        print(len(allConds))
        for ii in selection:
            trialList.append(allConds[int(ii)])

    logging.exp('Imported %s as conditions, %d conditions, %d params' %
                (fileName, len(trialList), len(fieldNames)))
    if returnFieldNames:
        return (trialList, fieldNames)
    else:
        return trialList


def createFactorialTrialList(factors):
    """Create a trialList by entering a list of factors with names (keys)
    and levels (values) it will return a trialList in which all factors
    have been factorially combined (so for example if there are two factors
    with 3 and 5 levels the trialList will be a list of 3*5 = 15, each
    specifying the values for a given trial

    Usage::

        trialList = createFactorialTrialList(factors)

    :Parameters:

        factors : a dictionary with names (keys) and levels (values) of the
            factors

    Example::

        factors={"text": ["red", "green", "blue"],
                 "letterColor": ["red", "green"],
                 "size": [0, 1]}
        mytrials = createFactorialTrialList(factors)
    """

    # the first step is to place all the factorial combinations in a list of
    # lists
    tempListOfLists = [[]]
    for key in factors:
        # this takes the levels of each factor as a set of values
        # (a list) at a time
        alist = factors[key]
        tempList = []
        for value in alist:
            # now we loop over the values in a given list,
            # and add each value of the other lists
            for iterList in tempListOfLists:
                tempList.append(iterList + [key, value])
        tempListOfLists = tempList

    # this second step is so we can return a list in the format of trialList
    trialList = []
    for atrial in tempListOfLists:
        keys = atrial[0::2]  # the even elements are keys
        values = atrial[1::2]  # the odd elements are values
        atrialDict = {}
        for i in range(len(keys)):
            # this combines the key with the value
            atrialDict[keys[i]] = values[i]
        # append one trial at a time to the final trialList
        trialList.append(atrialDict)

    return trialList


def bootStraps(dat, n=1):
    """Create a list of n bootstrapped resamples of the data

    SLOW IMPLEMENTATION (Python for-loop)

    Usage:
        ``out = bootStraps(dat, n=1)``

    Where:
        dat
            an NxM or 1xN array (each row is a different condition, each
            column is a different trial)
        n
            number of bootstrapped resamples to create

        out
            - dim[0]=conditions
            - dim[1]=trials
            - dim[2]=resamples
    """
    dat = np.asarray(dat)
    if len(dat.shape) == 1:
        # have presumably been given a series of data for one stimulus
        # adds a dimension (arraynow has shape (1,Ntrials))
        dat = np.array([dat])

    nTrials = dat.shape[1]
    # initialise a matrix to store output
    resamples = np.zeros(dat.shape + (n,), dat.dtype)
    rand = np.random.rand
    for stimulusN in range(dat.shape[0]):
        thisStim = dat[stimulusN, :]  # fetch data for this stimulus
        for sampleN in range(n):
            indices = np.floor(nTrials * rand(nTrials)).astype('i')
            resamples[stimulusN, :, sampleN] = np.take(thisStim, indices)
    return resamples


def functionFromStaircase(intensities, responses, bins=10):
    """Create a psychometric function by binning data from a staircase
    procedure. Although the default is 10 bins Jon now always uses 'unique'
    bins (fewer bins looks pretty but leads to errors in slope estimation)

    usage::

        intensity, meanCorrect, n = functionFromStaircase(intensities,
                                                          responses, bins)

    where:
            intensities
                are a list (or array) of intensities to be binned

            responses
                are a list of 0,1 each corresponding to the equivalent
                intensity value

            bins
                can be an integer (giving that number of bins) or 'unique'
                (each bin is made from aa data for exactly one intensity
                value)

            intensity
                a numpy array of intensity values (where each is the center
                of an intensity bin)

            meanCorrect
                a numpy array of mean % correct in each bin

            n
                a numpy array of number of responses contributing to each mean
    """
    # convert to arrays
    try:
        # concatenate if multidimensional
        intensities = np.concatenate(intensities)
        responses = np.concatenate(responses)
    except Exception:
        intensities = np.array(intensities)
        responses = np.array(responses)

    # sort the responses
    sort_ii = np.argsort(intensities)
    sortedInten = np.take(intensities, sort_ii)
    sortedResp = np.take(responses, sort_ii)

    binnedResp = []
    binnedInten = []
    nPoints = []
    if bins == 'unique':
        intensities = np.round(intensities, decimals=8)
        uniqueIntens = np.unique(intensities)
        for thisInten in uniqueIntens:
            theseResps = responses[intensities == thisInten]
            binnedInten.append(thisInten)
            binnedResp.append(np.mean(theseResps))
            nPoints.append(len(theseResps))
    else:
        pointsPerBin = len(intensities)/bins
        for binN in range(bins):
            start = int(round(binN * pointsPerBin))
            stop = int(round((binN + 1) * pointsPerBin))
            thisResp = sortedResp[start:stop]
            thisInten = sortedInten[start:stop]
            binnedResp.append(np.mean(thisResp))
            binnedInten.append(np.mean(thisInten))
            nPoints.append(len(thisInten))

    return binnedInten, binnedResp, nPoints


def getDateStr(format="%Y_%b_%d_%H%M"):
    """Uses ``time.strftime()``_ to generate a string of the form
    2012_Apr_19_1531 for 19th April 3.31pm, 2012.
    This is often useful appended to data filenames to provide unique names.
    To include the year: getDateStr(format="%Y_%b_%d_%H%M")
    returns '2011_Mar_16_1307' depending on locale, can have unicode chars
    in month names, so utf_8_decode them
    For date in the format of the current localization, do:
        data.getDateStr(format=locale.nl_langinfo(locale.D_T_FMT))
    """
    now = time.strftime(format, time.localtime())
    if PY3:
        return now
    else:
        try:
            now_decoded = codecs.utf_8_decode(now)[0]
        except UnicodeDecodeError:
            # '2011_03_16_1307'
            now_decoded = time.strftime("%Y_%m_%d_%H%M", time.localtime())

        return now_decoded
